"""This module is related to the parsing of meshtal files output from MCNP.

It includes parsing capabilities not only for classical MCNP fmesh, but also
for its variations introduced by the D1SUNED code, such as the Cell Under
Voxel (CuV) method.
"""
from __future__ import annotations
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/  
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import numpy as np
import vtk
import csv
import pyvista as pv
import os
import time
from scipy.spatial.transform import Rotation as R
from io import open
import logging
from tqdm import tqdm
from copy import deepcopy


ALLOWED_NORMALIZATIONS = ['vtot', 'celf', None]
ALLOWED_OUTPUT_FORMATS = ['point_cloud', 'ip_fluent', 'csv', 'vtk']


# convert character to float
# able to convert no standard fortran exponent type 1.234-123
def _dfloat(c: str) -> float:
    try:
        x = float(c)
    except:
        c = c.strip()
        m = float(c[0:-4])
        e = int(c[-4:])
        x = m * pow(10, e)
    return x


# return a list of each nth charaters of a string
def _splitn(string: str, n: int) -> list[str]:
    return [string[i: i + n].strip() for i in range(0, len(string), n)]


# Salta lineas (definida en cien sitios)
def _skipLines(f, n):
    for iskip in range(n):
        f.readline()
    return


# # read cell list from file for CUV filtering option
# def get_clist(fn: os.PathLike) -> list(int):
#     f = open(fn, "rt")
#     lines = f.readlines()
#     f.close()
#     clist = []
#     filter_mode = "accept"
#     for lin in lines:
#         line = lin.split("#")[0]

#         sc_num = line.count(":")
#         line = line.replace(":", " : ", sc_num)

#         val = line.split()
#         if "reject" in line:
#             filter_mode = "reject"
#             val.remove("reject")

#         scidx = 0
#         for i in range(sc_num):
#             sc = val.index(":", scidx)
#             low = int(val[sc - 1])
#             high = int(val[sc + 1])
#             del val[sc - 1: sc + 2]
#             scidx = sc
#             clist.extend(range(low, high + 1))

#         clist.extend(map(int, val))

#     clist.insert(0, filter_mode)
#     return clist


class Fmesh:
    dtype = np.float64
    cvarsCart = ("Z", "Y", "X")
    cvarsCyl = ("Theta", "Z", "R")
    IPT = ("neutron", "photon", "electron")

    # Reads from opened file
    def __init__(self, mshtl: Meshtal) -> None:
        """Object representing an Fmesh result.

        It needs a Meshtal object to be initialized. Once read, a
        PyVistaWrapper object will be assigned to it that can be used to 
        manipulate and export the parsed data.

        Parameters
        ----------
        mshtl : Meshtal
            meshtal object where the Fmesh is located

        Attributes
        ----------
        grid: pv.DataSet
            pyvista grid of the parsed data.
        meshtal: Meshtal
            the same meshtal object provided as parameter.
        comment: str
            comment string for the tally provided in MCNP input
        part: Literal['neutron', 'photon', 'electron']
            particle tallied in the fmesh.
        type: str
            type of fmesh, e.g. 'cuv', 'source', etc.
        normalization: str
            type of normalization used for the CuV approach
        filled : bool
            if True, the Fmesh has been populated with data.
        cart: bool
            if True the mesh is cartesian, otherwise cylindrical.
        ntally: int
            tally number for the fmesh

        Examples
        --------
        For examples check out :py:class:`Meshtal` doc.

        """
        self.filled = False
        self.ntally = -1
        self.part = "unknown"
        self.etag = "energy"
        self.type = "flux"
        self.dims = list()
        self.scaleFac = 1.0
        self.pos = 0
        self.startm = 0
        self.origin = np.zeros([4], self.dtype)
        self.meshtal = mshtl
        self.__format__ = "undef"
        self.comment = None
        self.dosecom = False
        self.usrbin = False
        self.readHead = False
        self.vec = None

        # here is stored the vtk object once the Fmesh is read
        self.grid = None

        # Rotation matrix
        self.rotation = np.identity(3, self.dtype)
        return

    def __readMeshCom__(self, f) -> None:
        line = f.readline()
        # read comment part
        if line[0:5] == "     ":
            self.comment = line
            line = f.readline()
            while line[0:5] == "     ":
                self.comment += line
                line = f.readline()
            self.comment = self.comment.strip("\n")

        # read particle type
        for p in ("neutron", "photon", "electron"):
            if p in line:
                self.part = p
                break

        if "importance mesh" in line:
            self.type = "srcimp"
        elif "cell-under-voxel" in line:
            self.type = "cuv"

        # Read MCNP mesh comments
        line = f.readline()
        while "Tally bin boundaries:" not in line:
            if "response function" in line:
                self.dosecom = True
            if "Energy bin" in line:
                if "energy" not in line[10:-1]:
                    self.usrbin = True
            if "source tally" in line:
                self.type = "source"
            if "Tally target:" in line:
                self.target = int(line.split()[-1])
                self.__format__ = "impmlt"
            line = f.readline()

        if self.type == "srcimp" and self.__format__ == "undef":
            self.__format__ == "impij"

        if self.type == "cuv":
            self.__format__ = "cuv"

    def __readMeshDim__(self, f) -> None:
        mcnp5_Cyl = "  Cylinder origin at"
        mcnp6_Cyl = "               origin at"

        # Test if cylinder
        self.cart = True
        line = f.readline()
        if line.startswith(mcnp5_Cyl) or line.startswith(mcnp6_Cyl):
            ishft = 0
            if line.startswith(mcnp6_Cyl):
                ishft = 1
            vals = line.split()
            self.cart = False
            # zval may end with comma
            zval = vals[5 - ishft]
            if not zval[-1].isdigit():
                zval = zval[:-1]
            self.origin = np.array(
                (0.0, zval, vals[4 - ishft], vals[3 - ishft]), self.dtype
            )
            self.axis = np.array(
                (vals[8 - ishft], vals[9 - ishft], vals[10 - ishft]),
                self.dtype
            )
            if line.startswith(mcnp6_Cyl):
                self.vec = np.array((vals[13], vals[14], vals[15]), self.dtype)
            line = f.readline()
        # Read dimensions  including energy
        # (E,Z,Y,X) order in storage in cartesian
        # (E,Th,Z,R) order in cylindrical
        for n in range(3):
            i = line.index(":")
            self.dims.insert(0, np.array(line[i + 1:].split(), self.dtype))
            line = f.readline()
        self.ldims = [len(x) - 1 for x in self.dims]

        # process energy bins
        i = line.index(":")
        if self.usrbin:
            # decay time bins
            if "times" in line:
                self.etag = "times"
                self.dims.insert(0, np.array(line[i + 1:].split(), self.dtype))
                self.ener = self.dims[0]  # shallow copy
                self.ldims.insert(0, len(self.dims[0]))
            # source importance bins
            elif "Target tally" in line:
                self.etag = "tally"
                line = f.readline()
                i = line.index(":")
                self.dims.insert(0, np.array(line[i + 1:].split(), self.dtype))
                self.ener = self.dims[0]  # shallow copy
                self.ldims.insert(0, len(self.dims[0]))
            # cell or isotopes bin
            else:
                self.etag = line[0:i].split()[0].lower()
                if self.etag == "cells":
                    binlen = 12
                else:
                    binlen = 10
                binlist = _splitn(line[i + 1:], binlen)[:-1]
                lastbin = binlist[-1]
                ib = lastbin.index(".")
                nbin = int(lastbin[ib + 1:])
                self.dims.insert(0, np.array(range(nbin), self.dtype))
                #        self.ener = np.array(line[i+1:].split(),self.dtype)
                self.ener = np.array(binlist, self.dtype)
                self.ldims.insert(0, nbin + 1)
        # energy bin
        else:
            if "number:" in line:
                i = line.index(":")
                ne = int(line[i + 1:])
                line = f.readline()
                i = line.index(":")
                if self.type not in ["srcimp", "cuv"]:
                    # check if onef or mltf
                    nelemts = (
                        self.ldims[0] * self.ldims[1] * self.ldims[2]
                    )  # dim of energy bin not yet inserted in ldims
                    # 6 columns 12 char each
                    # 2 block2 (flux, error)
                    nline = int(ne / 6)
                    nr = ne % 6
                    if nr != 0:
                        nline += 1
                    nblck = 2 * ((12 * ne) + nline)
                    pos = f.tell()
                    _skipLines(f, 5)
                    onef = _checkonef(f, nblck)
                    f.seek(pos)
                    if onef:
                        self.__format__ = "onef"
                    else:
                        self.__format__ = "mltf"

            self.dims.insert(0, np.array(line[i + 1:].split(), self.dtype))
            self.ener = self.dims[0]  # shallow copy
            nbin = 1
            if len(self.dims[0]) > 2:
                nbin = len(self.dims[0])
            self.ldims.insert(0, nbin)

        if self.__format__ == "undef":
            self.__format__ = "mcnp"
        self.startm = f.tell()

    def _readMCNP(self, f) -> None:
        """Read the MCNP file.

        Parameters
        ----------
        f : file stream
            file stream to be read
        """
        # Proceed to read MCNP
        # Check if file is readable
        # There is one blank between fmesh

        if self.readHead:
            f.seek(self.startm)
        else:
            line = f.readline()
            if line == "":
                print("read error")
            self.ntally = int(line[-6:-1])

            # read mesh comments header
            self.__readMeshCom__(f)

            # read mesh dimensions and bins
            self.__readMeshDim__(f)

        line = f.readline()
        # Coordinates are origin-independent
        if self.cart:
            for i in range(1, 4):
                self.origin[i] = self.dims[i][0]
                self.dims[i] -= self.origin[i]
        # Memory storage for cell data
        # Reversed indeces to keep ordering in memory
        rshape = self.ldims[:]
        self.dat = np.zeros(rshape, self.dtype)
        self.err = np.zeros(rshape, self.dtype)
        # Data ordering in file
        line = f.readline()
        matFormat = "Rel Error" not in line

        # select correct keyword depending on the user bin
        if self.etag == "energy":
            eword = "Energy"
        elif self.etag == "cells":
            eword = "Cell"
        elif self.etag == "times":
            eword = "Time"
        else:
            eword = "Nuclide"
        energyCol = eword in line

        # All meshtal orders have energy as the least moving index, so we will
        # read variable of one energy bin, and then transpose to proper
        # ordering transposing takes memory

        # Matrix format, what order?
        if matFormat:
            _skipLines(f, 1)
            if self.etag == "tally":
                self.tallyscore = [float(f.readline().split()[-1])]
            c = f.readline().split()[0]
            if self.cart:
                i = self.cvarsCart.index(c)
            else:
                i = self.cvarsCyl.index(c)
            if i == 0:  # ij
                iord = (1, 2, 3)  # Read order of storage
                itrn = (0, 1, 2)  # transpose of local storage
            elif i == 1:  # ik
                iord = (2, 1, 3)
                itrn = (1, 0, 2)
            elif i == 2:  # jk
                iord = (3, 1, 2)
                itrn = (1, 2, 0)
            else:
                print("Matrix transpose error")
        else:  # colformat x,y,z
            iord = (3, 2, 1)  # Z,Y,X para reshape
            itrn = (2, 1, 0)

        # Temporal variables for transposing
        # reshape of temp
        rshape = np.array([self.ldims[iord[i]] for i in range(3)])

        # ================================
        if matFormat:  # MATRIX FORMAT
            hlines = 3  # header lines to skip in first
            hlinesNext = 4  # Following hlines

            # Temporal variables for transposing
            xdat = np.zeros(rshape, self.dtype)
            xerr = np.zeros(rshape, self.dtype)

            for ie in range(self.ldims[0]):
                if ie > 0:
                    _skipLines(f, 3)
                    if self.etag == "tally":
                        self.tallyscore.append(float(f.readline().split()[-1]))

                for ix1 in range(rshape[0]):
                    _skipLines(f, hlines)
                    hlines = hlinesNext  # subsequent reads
                    for ix0 in range(rshape[1]):
                        xdat[ix1, ix0, :] = list(
                            map(_dfloat, f.readline().split()[1:])
                        )
                    # use dfloat conversion in case of no standar fortan
                    # exponent
                    _skipLines(f, 3)
                    for ix0 in range(rshape[1]):
                        xerr[ix1, ix0, :] = f.readline().split()[
                            1:
                        ]  # automatic conversion
                    _skipLines(f, 2)
                # Copy of temp data
                self.dat[ie, :, :, :] = np.transpose(xdat, itrn)
                self.err[ie, :, :, :] = np.transpose(xerr, itrn)

        else:  # colformat
            n = np.prod(rshape)
            xdat = np.zeros(n, self.dtype)
            xerr = np.zeros(n, self.dtype)
            colDat = 31
            colErr = 43
            if energyCol:  # first column is energy
                colDat += 10
                colErr += 10
            for ie in range(self.ldims[0]):  # energy
                for ix in range(n):
                    line = f.readline()
                    try:
                        xdat[ix] = _dfloat(line[colDat: colDat + 12])
                        xerr[ix] = line[colErr: colErr + 12]
                    except:
                        xdat[ix] = _dfloat(line[colDat + 1: colDat + 13])
                        xerr[ix] = line[colErr + 1: colErr + 13]
                self.dat[ie, :, :, :] = np.transpose(xdat.reshape(rshape),
                                                     itrn)
                self.err[ie, :, :, :] = np.transpose(xerr.reshape(rshape),
                                                     itrn)

        # Removal of thick angular bin if any
        # only if:
        # Angular domain is a whole revolution (there may be a thick bin in
        # between)
        if not self.cart and abs(self.dims[1][-1] - self.dims[1][0] - 1.0) < 0.01:
            maxDth = 5.0 / self.ldims[1]
            for i in range(self.ldims[1]):
                if self.dims[1][i + 1] - self.dims[1][i] > maxDth:
                    self.ldims[1] -= 1
                    np.delete(self.dat, i, 1)
                    np.delete(self.err, i, 1)
                    # Reorder remaining bins
                    if i > 0 and i < self.ldims[1] - 1:
                        self.dims[1] = np.concatenate(
                            (self.dims[1][i + 1: -1] - 1.0, self.dims[1][0:i]),
                             axis=0)
                        self.dat = np.concatenate(
                            (self.dat[:, i + 1:, :, :], self.dat[:, : i - 1, :, :]),
                            axis=1,
                        )
                        self.err = np.concatenate(
                            (self.err[:, i + 1:, :, :], self.err[:, : i - 1, :, :]),
                            axis=1,
                        )
                    else:
                        np.delete(self.dims[1], i, 0)
                    # Leave loop
                    break

        self.filled = True
        # if it is filled it means that the vtk object can be created
        name = '{}_{}'.format(self.meshtal.filename, self.ntally)
        if self.cart:
            self.grid = self._getVTKrg()
        else:
            self.grid = self._getVTKsg()

    # Read photonfile format of SRCIMP mesh (D1SUNED)
    def _readSRCTYPE(self, f, cfilter=None, norm=None):
        mcnp5_Cyl = "  Cylinder origin at"
        mcnp6_Cyl = "               origin at"
        self.etag = "energy"
        if norm is None:
            tag = '[/cc]'
        elif norm == 'vtot':
            tag = 'integral value'
        elif norm == 'celf':
            tag = '[/cc-cell]'
        else:
            tag = ''

        self.tag = tag  # to be used to clarify output array

        self.normalization = norm

        if self.readHead:
            f.seek(self.startm)
        else:
            line = f.readline()
            if line == "":
                print("Read error")
            self.ntally = int(line[-6:-1])

            # read mesh comments header
            self.__readMeshCom__(f)

            # read mesh dimensions and bins
            # erint("readDims")  # This is not defined!
            self.__readMeshDim__(f)

        # Coordinates are origin-independent
        if self.cart:
            for i in range(1, 4):
                self.origin[i] = self.dims[i][0]
                self.dims[i] -= self.origin[i]
        # Memory storage for cell data
        # Reversed indeces to keep ordering in memory
        if self.type == "cuv":
            _skipLines(f, 2)
        else:
            if self.type == "srcimp":
                self.tallyscore = np.float64(f.readline().split()[-1])
            _skipLines(f, 5)

        nelemts = self.ldims[1] * self.ldims[2] * self.ldims[3]
        nbin = int(self.ldims[0])

        if nbin > 1:
            ne = nbin - 1
        else:
            ne = nbin
        rshape = (nbin, nelemts)

        # 6 columns 12 char each
        # 2 block2 (flux, error)
        nline = int(ne / 6)
        nr = ne % 6
        if nr != 0:
            nline += 1

        # check if onef or mltf
        if self.__format__ == "undef":
            nblck = 2 * ((12 * ne) + nline)
            pos = f.tell()
            onef = _checkonef(f, nblck)
            f.seek(pos)
            if onef:
                self.__format__ = "onef"
            else:
                self.__format__ = "mltf"

        xdat = np.zeros(rshape, self.dtype)
        xerr = np.zeros(rshape, self.dtype)

        # source importance data format
        if self.type == "srcimp":
            for k in range(nelemts):
                vals = []
                errs = []
                tot = f.readline().split()[-1]
                for i in range(ne):
                    vals.extend(f.readline().split())
                for i in range(ne):
                    errs.extend(f.readline().split())

                if nbin > 1:
                    xdat[nbin - 1, k] = _dfloat(tot)
                    xerr[nbin - 1, k] = 0.0
                for i in range(ne):
                    xdat[i, k] = _dfloat(vals[i])
                    xerr[i, k] = float(errs[i])

        elif self.type == "cuv":
            # one/multiflux data format
            if cfilter is not None:
                if cfilter[0] == "reject":
                    accept = False
                else:
                    accept = True
            for k in range(nelemts):
                lvals = f.readline().split()
                nc = int(lvals[-1])
                voxvol = float(lvals[-2])

                celdata = [voxvol, [], []]
                cell = []

                # read cell data info
                for ic in range(nc):
                    vline = f.readline().split()
                    cline = int(vline[0])
                    cell.append(cline)
                    if (
                        cfilter is None
                        or (accept and cline in cfilter)
                        or (not accept and cline not in cfilter)
                    ):
                        celdata[1].append(float(vline[1]))
                        if nbin > 1:
                            celdata[2].append([float(vline[2]),
                                               float(vline[3])])

                voxvals = []
                voxerrs = []
                # read cell values and errors
                for ind, ic in enumerate(cell):
                    valc = []
                    errc = []
                    if (
                        cfilter is None
                        or (accept and ic in cfilter)
                        or (not accept and ic not in cfilter)
                    ):
                        for i in range(nline):
                            valc.extend(f.readline().split())
                        for i in range(nline):
                            errc.extend(f.readline().split())

                        vals = np.array([_dfloat(x) for x in valc], self.dtype)
                        errs = np.array([_dfloat(x) for x in errc], self.dtype)

                        if nbin > 1:
                            vals = np.append(vals, celdata[2][ind][0])
                            errs = np.append(errs, celdata[2][ind][1])

                        voxvals.append(vals)
                        voxerrs.append(errs)
                    else:
                        _skipLines(f, 2 * nline)

                if len(voxvals) == 0:
                    vals = [0] * nbin
                    errs = [0] * nbin
                else:
                    # sum all bin if not total bin
                    vals, errs = _sumCellInVox(
                        voxvals, voxerrs, celdata, Vmult=norm
                    )

                xdat[:, k] = vals[:]
                xerr[:, k] = errs[:]

        else:
            if cfilter is not None:
                if cfilter[0] == "reject":
                    accept = False
                else:
                    accept = True
            # one/multiflux data format
            for k in range(nelemts):
                lvals = f.readline().split()
                nc = int(lvals[-1])
                voxvol = float(lvals[-2])

                celfrac = [voxvol, []]
                cell = []
                for ic in range(nc):
                    vline = f.readline().split()
                    cline = int(vline[0])
                    cell.append(cline)
                    if (
                        cfilter is None
                        or (accept and cline in cfilter)
                        or (not accept and cline not in cfilter)
                    ):
                        celfrac[1].append(float(vline[1]))

                voxvals = []
                voxerrs = []

                if self.__format__ == "onef":
                    valc = []
                    errc = []
                    for i in range(nline):
                        valc.extend(f.readline().split())
                    for i in range(nline):
                        errc.extend(f.readline().split())

                    vals = np.array([_dfloat(x) for x in valc], self.dtype)
                    errs = np.array(errc, self.dtype)
                    voxvals.append(vals)
                    voxerrs.append(errs)
                else:
                    for ic in cell:
                        valc = []
                        errc = []
                        if (
                            cfilter is None
                            or (accept and cline in cfilter)
                            or (not accept and cline not in cfilter)
                        ):
                            for i in range(nline):
                                valc.extend(f.readline().split())
                            for i in range(nline):
                                errc.extend(f.readline().split())

                            vals = np.array([_dfloat(x) for x in valc],
                                            self.dtype)
                            errs = np.array(errc, self.dtype)
                            voxvals.append(vals)
                            voxerrs.append(errs)
                        else:
                            _skipLines(f, 2 * nline)

                if len(voxvals) == 0:
                    vals = [0] * nbin
                    errs = [0] * nbin
                    tot = 0
                else:
                    # sum all bin if not total bin
                    vals, errs = _sumElements(
                        voxvals,
                        voxerrs,
                        celfrac,
                        Vsum=self.__format__,
                        Vmult=norm,
                        corr=True,
                    )

                xdat[:, k] = vals[:]
                xerr[:, k] = errs[:]

        rshape = np.array([nbin, self.ldims[3], self.ldims[2], self.ldims[1]])
        self.dat = xdat.reshape(rshape)
        self.dat = self.dat.transpose(0, 3, 2, 1)
        self.err = xerr.reshape(rshape)
        self.err = self.err.transpose(0, 3, 2, 1)
        self.filled = True
        # if it is filled it means that the vtk object can be created
        # name = '{}_{}'.format(self.meshtal.filename, self.ntally)
        if self.cart:
            self.grid = self._getVTKrg()
        else:
            self.grid = self._getVTKrg()

    #  end modifs

    def print_info(self) -> None:
        """print information related to the fmesh
        """
        if self.__format__ == "mltf":
            meshtype = "multiflux"
        else:
            meshtype = self.type

        if self.cart:
            geom = "rectangular"
        else:
            geom = "cylindrical"

        print(" Tally          : {}".format(self.ntally))
        if self.comment is not None:
            print(" Comments       : ")
            print(" {}".format(self.comment))
        print(" Particle       : {}".format(self.part))
        if self.__format__ != "mltf":
            print(" Mesh type      : {}".format(meshtype))
        else:
            print(
                " Mesh type      : {}".format(meshtype)
            )
        print(" Dose modif     : {}".format(self.dosecom))
        print(" Mesh geometry  : {}".format(geom))
        print(" Mesh origin    : {org[3]} {org[2]} {org[1]}".format(org=self.origin))
        if self.cart:
            print(
                " X dimensions   :{}".format(
                    self.__format_XYZ_Dim__(self.dims[3] + self.origin[3])
                )
            )
            print(
                " Y dimensions   :{}".format(
                    self.__format_XYZ_Dim__(self.dims[2] + self.origin[2])
                )
            )
            print(
                " Z dimensions   :{}".format(
                    self.__format_XYZ_Dim__(self.dims[1] + self.origin[1])
                )
            )
        else:
            print(" R dimensions   :{}".format(self.__format_XYZ_Dim__(self.dims[3])))
            print(" Z dimensions   :{}".format(self.__format_XYZ_Dim__(self.dims[2])))
            print(" T dimensions   :{}".format(self.__format_XYZ_Dim__(self.dims[1])))
        print(" Energy bins    :")
        self.print_EbinRange()

    def __format_XYZ_Dim__(self, vec: list, nval: int = 6) -> str:
        def format_XYZ_Dim_inter(vec, nval=6):
            interfound = False
            xtab = []
            dx0 = 0
            x1 = vec[0]
            i = 0
            for x2 in vec[1:]:
                dx = round(x2 - x1, 2)
                if dx == dx0:
                    i += 1
                    interfound = True
                else:
                    xtab.append([i, x1])
                    dx0 = dx
                    i = 0
                x1 = x2
            xtab.append([i, x2])

            if not interfound:
                return _format_XYZ_Dim_long(vec)
            line = ""
            newline = "  {:10.3e}".format(xtab[0][1])

            for i, v in enumerate(xtab[1:]):
                newline += " {val[0]:4d}I {val[1]:10.3e}".format(val=v)
                if (i + 1) % nval == 0:
                    line += newline + "\n"
                    newline = "                 "

            if (i + 1) % nval != 0:
                line += newline
            else:
                line = line[:-1]
            return line

        def _format_XYZ_Dim_long(vec: list, nval: int = 8) -> str:
            line = ""
            newline = ""
            for i, v in enumerate(vec):
                newline += " {:10.3e}".format(v)
                if (i + 1) % nval == 0:
                    line += newline + "\n"
                    newline = "                 "

            if (i + 1) % nval != 0:
                line += newline
            else:
                line = line[:-1]
            return line

        if len(vec) <= nval:
            return _format_XYZ_Dim_long(vec)
        else:
            return format_XYZ_Dim_inter(vec)

    def print_EbinRange(self) -> None:
        """print the energy range bins
        """
        print("         flag : {}".format(self.etag))
        print("    bin index : bin range  ")
        if self.etag == "energy":
            nb = 0
            if len(self.ener) > 2:
                nb = len(self.ener) - 1
                for i in range(nb):
                    print(
                        "       {:4d}   :      {} - {}  MeV".format(
                            i, self.ener[i], self.ener[i + 1]
                        )
                    )

        elif self.etag in ["cells", "daughter", "parent"]:
            binlist = []
            ibin = 0
            for b in self.ener:
                decpart, intpart = np.modf(b)
                binnum = int(np.rint(abs(decpart) * 1000)) - 1
                if binnum == ibin:
                    binlist.append([int(intpart)])
                    ibin += 1
                else:
                    binlist[binnum].append(int(intpart))

            ib = 0
            print("       {:4d}   :  Other ".format(ib))
            for b in binlist[1:]:
                ib += 1
                line = "       {:4d}   :".format(ib)
                rang = 0
                for i, v in enumerate(b):
                    if i < len(b) - 1:
                        if b[i + 1] < 0:
                            rang = 1
                    if rang == 0:
                        line = line + " {},".format(v)
                    elif rang == 1:
                        line = line + " {}-".format(v)
                        rang = 2
                    else:
                        line = line + "{},".format(abs(v))
                        rang = 0
                print(line[:-1])

            ib += 1
            print("       {:4d}   :  Total ".format(ib))

        elif self.etag in ["times", "tally"]:
            for i, b in enumerate(self.ener):
                print("       {:4d}   :      {}".format(i, b))
        return

    # Checks whether it is the same mesh
    def sameMesh(self, xelf, checkErg: bool = False) -> bool:
        i1 = 1
        if checkErg:
            i1 = 0
            if self.etag != xelf.etag:
                return False

        if self.cart != xelf.cart:
            return False
        if self.ldims[i1:4] != xelf.ldims[i1:4]:
            return False
        if np.linalg.norm(self.origin - xelf.origin) > 0.001:
            return False
        if any(
            (np.linalg.norm(self.dims[i] - xelf.dims[i]) > 0.01 for i in range(i1, 4))
        ):
            return False
        if not self.cart:
            if np.linalg.norm(self.axis - xelf.axis) > 0.001:
                return False
            if self.vec is not None:
                if np.linalg.norm(self.vec - xelf.vec) > 0.001:
                    return False

        return True

    # # Translate the mesh
    # def translate(self, vec: list[float]) -> None:
    #     """translate the fmesh using a vector x, y, z.

    #     Parameters
    #     ----------
    #     vec : list[float]
    #         translation vector
    #     """
    #     self.origin += np.array(vec)

    # Salida en formato VTK
    # Escribe structured grid
    def _getVTKsg(self) -> pv.DataSet:
        """get the VTK structured grid format for the mesh

        Returns
        -------
        pv.DataSet
            generated vtk object
        """
        import math

        kdims = self.dims
        pts = vtk.vtkPoints()
        pts.SetDataTypeToFloat()

        # Paso de celdas degeneradas en el centro
        # Nunca se considera el centro del cilindro
        if not self.cart and kdims[3][0] == 0.0:
            kdims[3][0] = kdims[3][1] * 0.01

        # Add an angular bin if first bin angle >0.5
        if not self.cart and kdims[1][1] > 0.5:
            self.ldims[1] += 1
            kdims[1] = np.insert(kdims[1], 1, 0.5, axis=0)
            self.dat = np.insert(self.dat, 0, self.dat[:, 0, :, :], axis=1)
            self.err = np.insert(self.err, 0, self.err[:, 0, :, :], axis=1)

        ptDims = np.array([len(kdims[3]), len(kdims[2]), len(kdims[1])])
        pts.SetNumberOfPoints(np.prod(ptDims))

        ps = []
        k = 0
        dpt = self.origin[:0:-1]  # origin en reversa
        for z in kdims[1]:
            if not self.cart:
                ang = 2.0 * math.pi * z
                sz = math.sin(ang)
                cz = math.cos(ang)
            for y in kdims[2]:
                for x in kdims[3]:
                    if self.cart:
                        pt = np.array((x, y, z))
                    else:
                        pt = np.array((x * cz, x * sz, y))  # por ahora axs = 0 0 1
                    ps.append(pt + dpt)
                    k += 1

        # cylindrical rotation
        ps -= dpt
        rotAxis = np.cross(self.axis, [0.0, 0.0, 1.0])
        if np.linalg.norm(rotAxis) == 0:
            rotM = R.from_rotvec([0, 0, 0])
        else:
            rotAxis = rotAxis / np.linalg.norm(rotAxis)
            ma = np.linalg.norm(self.axis)
            ang = -np.arccos(np.dot(self.axis, [0.0, 0.0, 1.0]) / ma)
            rotM = R.from_rotvec(rotAxis * ang)
        ps = rotM.apply(ps)
        ps += dpt
        for p in range(len(ps)):
            pts.InsertPoint(p, ps[p])

        # Puntos definidos
        sg = vtk.vtkStructuredGrid()
        sg.SetDimensions(*ptDims)

        sg.SetPoints(pts)
        sgcd = sg.GetCellData()
        # These slicing operations do not copy data
        # mySlice = [slice(-1,-2,-1),slice(None),slice(None),slice(None)]
        if self.__format__ == 'cuv':
            value_tag = 'Value '+self.tag
        else:
            value_tag = 'Value - Total'

        it = 0
        if self.etag not in ["times", "tally"]:
            # Dataset energia total
            it = 1
            sgcd.AddArray(
                _makeVTKarray(self.dat[-1, :, :, :], value_tag, self.scaleFac)
            )
            sgcd.AddArray(_makeVTKarray(self.err[-1, :, :, :],
                                        "Error - Total"))
        # Dataset other bins
        for ie in range(self.ldims[0] - it):
            sgcd.AddArray(
                _makeVTKarray(
                    self.dat[ie, :, :, :], "ValueBin-{0:03d}".format(ie),
                    self.scaleFac
                )
            )
            sgcd.AddArray(
                _makeVTKarray(self.err[ie, :, :, :],
                              "ErrorBin-{0:03d}".format(ie))
            )
        # Include weight windows in VTK file (now only one group)

        # TODO use pyvista to create the object
        # for the moment just wrap in a pyvista object
        return pv.wrap(sg)

    # def _writeVTKsg(self, ofn: os.PathLike) -> None:
    #     """Write the mesh to a vtk file

    #     Parameters
    #     ----------
    #     ofn : os.PathLike
    #         outpath for the file
    #     """
    #     try:
    #         name = 'tally_{}_{}.vts'.format(self.ntally, self.tag)
    #     except AttributeError:
    #         name = 'tally_{}.vts'.format(self.ntally)

    #     # Escritura en disco
    #     off = vtk.vtkXMLStructuredGridWriter()
    #     off.SetFileName(os.path.join(ofn, name))
    #     # ASCII o binario (con o sin compresion)
    #     off.SetDataModeToAscii()
    #     # off.SetDataModeToBinary()
    #     # off.SetCompressorTypeToZLib()
    #     # Esto cambia con la version de VTK
    #     t = self.getVTKsg()
    #     self.meshtal._setVTKparams(t)
    #     if vtk.vtkVersion().GetVTKMajorVersion() >= 6:
    #         off.SetInputData(t)
    #     else:
    #         off.SetInput(t)
    #     off.Write()

    # Escribe rectilinear grid
    # No vale para cilindricas o rotadas
    def _getVTKrg(self) -> pv.DataSet:
        """Get a rectilinear grid. This cannot be used for cylindrical or
        rotated grids

        Returns
        -------
        pv.DataSet
            generated rectilinear grid object
        """
        if not self.cart:
            logging.warning(
                "Cylindrical meshtal cannot be plotted to RectangularGrid")

        if np.any(self.rotation != np.identity(3)):
            logging.warning(
                "Rotated meshtal cannot be plotted to RectangularGrid")
            logging.warning("... but plotting anyway (no rotation)")

        xa = _makeVTKarray(self.dims[3] + self.origin[3], "X (cm)")
        ya = _makeVTKarray(self.dims[2] + self.origin[2], "Y (cm)")
        za = _makeVTKarray(self.dims[1] + self.origin[1], "Z (cm)")

        rg = vtk.vtkRectilinearGrid()
        rg.SetDimensions(self.ldims[3] + 1, self.ldims[2] + 1,
                         self.ldims[1] + 1)
        rg.SetXCoordinates(xa)
        rg.SetYCoordinates(ya)
        rg.SetZCoordinates(za)
        rgcd = rg.GetCellData()
        it = 0
        if self.__format__ == 'cuv':
            value_tag = 'Value '+self.tag
        else:
            value_tag = 'Value - Total'

        if self.etag not in ["times", "tally"]:
            # Dataset energia total
            it = 1
            rgcd.AddArray(
                _makeVTKarray(self.dat[-1, :, :, :], value_tag,
                              self.scaleFac)
            )
            rgcd.AddArray(_makeVTKarray(self.err[-1, :, :, :], "Error - Total"))
        # Dataset other bins
        for ie in range(self.ldims[0] - it):
            rgcd.AddArray(
                _makeVTKarray(
                    self.dat[ie, :, :, :], "ValueBin-{0:03d}".format(ie),
                    self.scaleFac
                )
            )
            rgcd.AddArray(
                _makeVTKarray(self.err[ie, :, :, :],
                              "ErrorBin-{0:03d}".format(ie))
            )

        # TODO use pyvista to create the object
        # for the moment just wrap in a pyvista object
        return pv.wrap(rg)

    # def _writeVTKrg(self, ofn: os.PathLike) -> None:
    #     """Write the rectilinear grid

    #     Parameters
    #     ----------
    #     ofn : os.PathLike
    #         output vtk file path
    #     """
    #     try:
    #         name = 'tally_{}_{}.vtr'.format(self.ntally, self.tag)
    #     except AttributeError:
    #         name = 'tally_{}.vtr'.format(self.ntally)
    #     # Escritura en disco
    #     off = vtk.vtkXMLRectilinearGridWriter()
    #     off.SetFileName(os.path.join(ofn, name))
    #     off.SetDataModeToAscii()
    #     # Esto cambia con la version de VTK
    #     t = self._getVTKrg()
    #     self.meshtal._setVTKparams(t)
    #     if vtk.vtkVersion().GetVTKMajorVersion() >= 6:
    #         off.SetInputData(t)
    #     else:
    #         off.SetInput(t)
    #     off.Write()

    # def writeVTK(self, ofn: os.PathLike) -> None:
    #     """Write the fmesh to a vtk file

    #     Parameters
    #     ----------
    #     ofn : os.PathLike
    #         path to the vtk outfile folder
    #     """
    #     if not os.path.exists(ofn):
    #         raise IsADirectoryError('Directory does not exists {}'.format(ofn))

    #     try:
    #         name = 'tally_{}_{}'.format(self.ntally, self.tag)
    #     except AttributeError:
    #         name = 'tally_{}'.format(self.ntally)

    #     if self.cart:
    #         dataset = self._getVTKrg()
    #     else:
    #         dataset = self._getVTKrg()

    def write(self, outpath: os.PathLike,
              list_array_names: list[str] = None,
              out_format: str = 'vtk',
              outfile: str = None) -> None:
        """Export the mesh to a file. vtk, csv, fluent (txt) and point cloud
        (txt) formats can be selected.

        Parameters
        ----------
        outpath : os.PathLike
            path to the output folder.
        list_array_names : list[str], optional
            arrays to be exported. The default is None, meaning that all the 
            available arrays will be used.
        out_format : str, optional
            output format. The allowed ones are ['point_cloud', 'ip_fluent',
            'csv', 'vtk']. Default is .vtk
        outfile : str, optional
            name of the output file. If specified, overrides the default one.
            Do not include the extension of the file here. Default is None.

        Raises
        ------
        KeyError
            raises KeyError if the output format is not allowed.
        """
        if list_array_names is None:
            list_array_names = list(self.grid.array_names)

        if outfile is None:
            file_name = f"{self.meshtal.filename}_{self.ntally}_{out_format}"
        else:
            file_name = outfile

        # TODO either all cells or all point data are supported if not a vtk
        if out_format != 'vtk':
            len_data = len(self.grid.cell_data)
            len_point = len(self.grid.point_data)
            if len_data > 0 and len_point == 0:
                f_points = self.grid.cell_centers().points
            elif len_point > 0:
                f_points = self.grid.points
            else:
                raise ValueError(
                    'mix between cell and point data is only supported for vtk'
                    )

        filepath = os.path.join(outpath, file_name)
        mesh_type = str(type(self.grid)).split(".")[-1][:-2]

        if out_format == 'vtk':
            if mesh_type == "StructuredGrid":
                ext = '.vts'
            elif mesh_type == "UnstructuredGrid":
                ext = '.vtu'
            elif mesh_type == "RectilinearGrid":
                ext = '.vtr'
            else:
                ext = '.vtk'

            self.grid.save(filepath+ext)
            return

        # --- CSV writer ---
        elif out_format == "csv":
            new_name = filepath + '.csv'

            with open(new_name, "w", newline="") as outfile:
                writer = csv.writer(outfile)

                # # TODO This may create some issues...
                # values_type = self.get_array_type(list_array_names[0])
                # if values_type == "cells":  # Take points or centers
                #     f_points = self.centers
                # else:  # Points
                #     f_points = self.points

                for i in tqdm(range(len(f_points)),
                              unit=" Points", desc="Writing"):
                    csv_points = [
                        f"{f_points[i][0]:.3f}",
                        f" {f_points[i][1]:.3f}",
                        f" {f_points[i][2]:.3f}",
                    ]
                    for array_name in list_array_names:
                        csv_points.append(
                            f" {self.grid[array_name][i]:.3f}")
                    writer.writerow(csv_points)

                logging.info(f"{new_name} created successfully!")
            return

        for array_name in list_array_names:
            # values_type = self.get_array_type(list_array_names[0])
            # if values_type == "cells":  # Take points or centers
            #     f_points = self.centers
            # else:  # Points
            #     f_points = self.points
            values = self.grid[array_name]

            # write depending on format
            # --- point cloud writer ---
            if out_format == "point_cloud":
                new_name = filepath + '.txt'
                with open(new_name, "w") as outfile:
                    outfile.write("x, y, z, value\n")
                    # TODO this can probably be optmized using
                    # outfile.writeline()
                    for i in tqdm(range(len(f_points)),
                                  unit=" Points", desc="Writing"):
                        outfile.write(f"{f_points[i][0]:.3f},")
                        outfile.write(f"{f_points[i][1]:.3f},")
                        outfile.write(f"{f_points[i][2]:.3f},")
                        outfile.write(f"{values[i]:.3f}\n")
                logging.info(f"{new_name} created successfully!")
                return

            # --- fluent writer ---
            elif out_format == "ip_fluent":
                new_name = filepath + '.txt'

                with open(new_name, "w") as outfile:
                    guion1 = "3"
                    n_coord = f_points.shape[1]  # self.n_coordinates
                    n_values = str(len(f_points))
                    guion2 = "1"
                    uds = "uds-0"
                    beginning = f"{guion1}\n{n_coord}\n{n_values}\n{guion2}\n{uds}\n"
                    outfile.write(beginning)
                    outfile.write("(")
                    for i in tqdm(range(len(f_points)),
                                  unit=" x points", desc="Writing x"):
                        outfile.write(f"{f_points[i][0]:.3f}\n")

                    outfile.write(")\n")
                    outfile.write("(")

                    for i in tqdm(range(len(f_points)),
                                  unit=" y points", desc="Writing y"):
                        outfile.write(f"{f_points[i][1]:.3f}\n")

                    outfile.write(")\n")
                    outfile.write("(")

                    for i in tqdm(range(len(f_points)),
                                  unit=" z points", desc="Writing z"):
                        outfile.write(f"{f_points[i][2]:.3f}\n")

                    outfile.write(")\n")
                    outfile.write("(")

                    for i in tqdm(range(len(f_points)),
                                  unit=" values", desc="Writing values"):
                        outfile.write(f"{values[i]:.3f}\n")

                    outfile.write(")\n")

                logging.info(f"{new_name} created successfully!")
                return

        raise KeyError(
            "Invalid format, these are the ones allowed: {}".format(ALLOWED_OUTPUT_FORMATS))

    def _read_from_vtk(self, vtk_file: os.PathLike):
        # This is mostly used for quicker testing
        grid = pv.read(vtk_file)
        self.grid = grid


class Meshtal:
    def __init__(self, fn: os.PathLike, filetype: str = "MCNP") -> None:
        """Class representing a parsed Meshtal file

        Parameters
        ----------
        fn : os.PathLike
            path to the meshtal file to be read
        filetype : str, optional
            type of file, by default "MCNP" which is the only file type
            implemented at the moment.

        Attributes
        ----------
        filename: str
            name of the file which is extracted by the fn path
        filetype: str
            same as the parameter
        mesh: dict[str, Fmesh]
            a dictionary containing all the Fmesh objects that are read from
            the meshtal file

        Examples
        --------
        Examples of usage of the Meshtal object

        >>> from f4enix.output.meshtal import Meshtal
        ... # Initialize the meshtal file
        ... file = 'cuvmsh'
        ... meshtal = Meshtal(file)
        ... print(meshtal)
        Meshtally file : cuvmsh   Tally 24 : neutron  cuv mesh   'Test cell under voxel' 

        and then read all meshes available and access their associated
        PyVistaWrapper object

        >>> # if no mesh are specified all mesh are read
        ... meshtal.readMesh()
        ... # Once the mesh is read, the grid attribute gets filled
        ... print(type(meshtal.mesh[24].grid))
        ... meshtal.mesh[24].grid
        {24: <f4enix.output.meshtal.Fmesh at 0x1badb9d4310>}
        ...
        <class 'pyvista.core.grid.RectilinearGrid'>
        ...
        Name: cuvmsh_24
        ...
        Header	Data Arrays
        RectilinearGrid	Information
        N Cells	2293504
        N Points	2346125
        X Bounds	-1.700e+03, 1.700e+03
        Y Bounds	-1.700e+03, 1.700e+03
        Z Bounds	-1.360e+03, 1.740e+03
        Dimensions	137, 137, 125
        N Arrays	10
        Name	        Field	Type	N   	Min	       Max
        Value - Total	Cells	float64	1	0.000e+00	5.286e+08
        Error - Total	Cells	float64	1	0.000e+00	1.000e+00
        ValueBin-000	Cells	float64	1	0.000e+00	1.547e+05
        ErrorBin-000	Cells	float64	1	0.000e+00	1.000e+00
        ValueBin-001	Cells	float64	1	0.000e+00	3.497e+08
        ErrorBin-001	Cells	float64	1	0.000e+00	1.000e+00
        ValueBin-002	Cells	float64	1	0.000e+00	1.787e+08
        ErrorBin-002	Cells	float64	1	0.000e+00	1.000e+00
        ValueBin-003	Cells	float64	1	0.000e+00	0.000e+00
        ErrorBin-003	Cells	float64	1	0.000e+00	0.000e+00

        Cell Under Voxel (CuV) approach is supported where cell_filters and
        different normalization can be chosen. The fmesh can be exported to
        different formats

        >>> # Read a specific CuV fmesh filtering by a set of cells
        ... # and changing the normalization to get the integral result
        ... meshtal.readMesh(24, cell_filters=[1], norm='vtot')
        ... fmesh = meshtal.mesh[24]
        ... fmesh.print_info()
        ... outpath = 'folderpath'
        ... # it can be exported to different formats, extension will be
        ... # inferred automatically
        ... fmesh.write(outpath)
        ... fmesh.write(outpath, out_format='ip_fluent')
        Tally          : 24
        Comments       : 
            Test cell under voxel                                                      
        Particle       : neutron
        Mesh type      : cuv
        Dose modif     : False
        Mesh geometry  : rectangular
        Mesh origin    : 0.0 0.0 0.0
        X dimensions   :   0.000e+00   50I  1.020e+02
        Y dimensions   :   0.000e+00   50I  1.020e+02
        Z dimensions   :   0.000e+00   50I  1.020e+02
        Energy bins    :
                flag : energy
            bin index : bin range  
        Writing x: 100%|██████████| 132651/132651 [00:00<00:00, 204617.39 x points/s]
        Writing y: 100%|██████████| 132651/132651 [00:00<00:00, 217292.19 y points/s]
        Writing z: 100%|██████████| 132651/132651 [00:00<00:00, 227737.30 z points/s]
        Writing values: 100%|██████████| 132651/132651 [00:00<00:00, 770635.26 values/s]

        """
        logging.info('Loading Meshtal: {}'.format(fn))
        self.filename = os.path.basename(fn).split('.')[0]
        # Parametros para VTK
        self.filetype = filetype
        self.f = open(fn, "rt")
        self.__readHeadMCNP__()
        self.mesh = self.__scanMCNP__()

        self.params = dict()
        self.params["creationTime"] = time.asctime()
        if "USER" in os.environ:
            self.params["author"] = os.environ["USER"]
        elif "USERNAME" in os.environ:
            self.params["author"] = os.environ["USERNAME"]
        if "HOSTNAME" in os.environ:
            self.params["host"] = os.environ["HOSTNAME"]
        elif "COMPUTERNAME" in os.environ:
            self.params["host"] = os.environ["COMPUTERNAME"][2:]
        self.params["path"] = os.getcwd()
        return

    def __scanMCNP__(self) -> dict[int, Fmesh]:
        mcnp5_Cyl = "  Cylinder origin at"
        mcnp6_Cyl = "               origin at"
        mesh = {}
        while True:
            line = self.f.readline()
            if line == "":
                break
            if "Mesh Tally Number" in line:
                t = Fmesh(self)
                ntally = int(line.split()[-1])
                t.ntally = ntally
                t.pos = self.f.tell() - len(line)

                t.__readMeshCom__(self.f)
                t.__readMeshDim__(self.f)
                t.readHead = True

                mesh[ntally] = t
        return mesh

    def readMesh(self, mesh: int | list[int] = None,
                 cell_filters: list[int] = None,
                 norm: str = None) -> None:
        """Parse a list of FMESHes

        Parameters
        ----------
        mesh : int | list[int], optional
            list of FMESH to be read, by default None means that all meshes
            will be parsed
        cell_filters : list[int], optional
            list of cells to be used as filters for the CuV approach,
            by default None
        norm : str
            to be used only on CuV. Can be set either to "vtot" or "celf".
            Default is None which means that the normalzation will be done
            by dividing by the Voxel volume. 'vtot' means that the default
            (None) will be multiplied by the total voxel volume to obtain an
            integral value, while "celf" normalizes dividing by the volume of
            the cell fraction under the voxel.

        Raises
        ------
        KeyError
            if the filetype is is not implemented
        """
        logging.info('Reading fmesh: {}'.format(mesh))
        if norm not in ALLOWED_NORMALIZATIONS:
            raise NameError(
                '{} is not an allowed normalization keyword'.format(norm))
        if self.filetype == "MCNP":
            # cycle on all meshes to be read
            if mesh is None:
                mesh = self.mesh.keys()
            elif type(mesh) is int:
                mesh = [mesh]

            for meshid in mesh:
                m = self.mesh[meshid]
                # move to the mesh location in the file
                self.f.seek(m.pos)
                if m.__format__ == "mcnp":
                    m._readMCNP(self.f)
                else:
                    # if m.__mltflt__ is not None:
                    #     flist = get_clist(m.__mltflt__)
                    # else:
                    #     flist = None
                    m._readSRCTYPE(self.f, cfilter=cell_filters, norm=norm)
        else:
            raise KeyError("This file type is not implemented: " +
                           str(self.filetype))

    def write_all(self, outpath: os.PathLike, out_format: str = 'vtk') -> None:
        """write all fmeshes to the outfolder in the specified format.

        Parameters
        ----------
        outpath : os.PathLike
            path to the output folder.
        out_format : str, optional
            output format. The allowed ones are ['point_cloud', 'ip_fluent',
            'csv', 'vtk']. Default is .vtk

        """
        for _, mesh in self.mesh.items():
            mesh.write(outpath, out_format=out_format)

    def __readHeadMCNP__(self) -> None:
        vals = self.f.readline().split()
        self.code = vals[0]
        self.version = vals[2]
        self.probid = vals[-2] + " " + vals[-1]
        self.title = self.f.readline().strip()
        self.nps = int(
            float((self.f.readline().split()[-1]))
        )  # nps: int doesnt like decimals

    def collapse_grids(self, name_dict: dict[int, list[str, str]]
                       ) -> pv.DataSet:
        """If the all the fmeshes in the meshtal are defined on the same
        structured grid, returns a grid onto which all the fmeshes are
        collapsed. That is, the returned grid will have all the field data that
        are stored in the different fmeshes.

        Parameters
        ----------
        name_dict : dict[int, list[str, str]]
            this dictionary is used to assign names to the fields that are
            added to the grid. The key of the dictionary is the number of the
            fmesh, while the list of strings are
            ['name of the values', 'name of the values statistical error'].
            For instance {104: ['neutron heating [W/cc]',
            'neutron heating [Rel err]']}

        Returns
        -------
        pv.DataSet
            returns the pyvista grid where all fields from the different
            Fmeshes have been added

        Raises
        ------
        RuntimeError
            if the fmeshes have different geometry if they do not contain
            only the default fields ['Value - Total', 'Error - Total'].
        """

        ids = ['Value - Total', 'Error - Total']

        try:
            # check that the collapse is doable
            for i, (_, fmesh) in enumerate(self.mesh.items()):
                # Check they are all same size
                if i == 0:
                    try:
                        comparison = fmesh
                    except AttributeError:
                        # it means that the mesh has not been read
                        raise AttributeError(
                            'Please read the meshtal file first using readMesh()')
                else:
                    assert comparison.sameMesh(fmesh)
        except AssertionError:
            raise RuntimeError('the different fmeshes geom are not compatible')

        try:
            # check that the collapse is doable
            for i, (_, fmesh) in enumerate(self.mesh.items()):
                # check that there are only the two usual values, no binning
                assert fmesh.grid.array_names == ids
        except AssertionError:
            raise RuntimeError('no binning allowed for the collapse')

        for i, (key, fmesh) in enumerate(self.mesh.items()):
            if i == 0:
                grid = deepcopy(fmesh.grid)
                for old_name, new_name in zip(grid.array_names,
                                              name_dict[key]):
                    grid.rename_array(old_name, new_name)
            else:
                # results just have to be added for the other fmeshes
                for array_name, id in zip(name_dict[key], ids):
                    grid[array_name] = fmesh.grid[id]

        return grid

    # def writeVTK(self, ofn: os.PathLike) -> None:
    #     """write a .vtk file for all read fmeshes

    #     Parameters
    #     ----------
    #     ofn : os.PathLike
    #         path to the output folder
    #     """

    #     mb = vtk.vtkMultiBlockDataSet()
    #     i = 0
    #     for k, m in self.mesh.items():
    #         if not m.filled:
    #             continue
    #         if m.cart:
    #             mb.SetBlock(i, m._getVTKrg())
    #             extension = '.vtr'
    #         else:
    #             mb.SetBlock(i, m._getVTKsg())
    #             extension = '.vts'
    #         i += 1
    #     # A escribir
    #     outpath = os.path.join(ofn, self.filename+extension)
    #     off = vtk.vtkXMLMultiBlockDataWriter()
    #     self._setVTKparams(mb)
    #     off.SetInputData(mb)
    #     off.SetFileName(outpath)
    #     off.Write()

    # def _addVTKparams(self, xdict: dict) -> None:
    #     """Add more parameters to include in VTK

    #     Parameters
    #     ----------
    #     xdict : dict
    #         parameters to be added
    #     """
    #     self.params.update(xdict)

    # def _setVTKparams(self, xdat: vtk.vtkDataSet) -> None:
    #     """Modify VTK structure for including parameters

    #     Parameters
    #     ----------
    #     xdat : vtk.vtkDataSet
    #         vtk data set to be modified
    #     """
    #     # for x,v in self.params.iteritems():
    #     for x, v in self.params.items():
    #         t = vtk.vtkStringArray()
    #         t.SetName(x)
    #         t.InsertNextValue(v)
    #         xdat.GetFieldData().AddArray(t)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        out = ''
        out = out + "\n Meshtally file : {}".format(self.filename)
        # tlist = self.mesh.keys()
        tlist = list(self.mesh.keys())
        tlist.sort()
        for t in tlist:
            m = self.mesh[t]
            if m.__format__ == "mltf":
                meshtype = "multiflux"
            else:
                meshtype = m.type
            if m.comment is not None:
                firstlinecom = m.comment.split("\n")[0].strip()
                out = out + (
                    "   Tally {} : {}  {} mesh   '{}' ".format(
                        t, m.part, meshtype, firstlinecom
                    )
                )
            else:
                out = out + (
                    "   Tally {} : {}  {} mesh ".format(t, m.part, meshtype))
        return out

    def print_info(self) -> None:
        """print information on the Meshtal
        """
        print(self)


# ================= END OF CLASS DEFINITIONS ======================


# This should be the same as above
def _makeVTKarray(nArr, aName, sc=1.0):
    from vtk.util import numpy_support

    if sc == 1.0:
        vArr = numpy_support.numpy_to_vtk(nArr.ravel(), deep=1)
    else:
        # deep copy needed in both cases. Seems the multiplication does not
        # create space
        vArr = numpy_support.numpy_to_vtk(nArr.ravel() * sc, deep=1)
    vArr.SetName(aName)
    return vArr


# check if flux mesh from mcnpacab is onef or mltf format
def _checkonef(f, lblck):
    ncel = 0
    while True:
        vals = f.readline().split()
        pos = f.tell()
        nvals = len(vals)
        if ncel > 1 and nvals == 9:
            return True

        if nvals == 9:
            ncel = int(vals[-1])
            pos = pos + 19 * ncel + lblck
            f.seek(pos)
        else:
            if nvals == 0:
                return True
            else:
                return False


def _sumCellInVox(voxvals, voxerrs, celfrac, Vmult="none", corr=True,
                  nulcount=True):
    # sum the value in the voxel multiplied by the volume fraction
    valf = list(map(lambda x, y: x * y, voxvals, celfrac[1]))

    if nulcount:
        volOK = sum(celfrac[1])
    else:
        volOK = 0
        for i, x in enumerate(valf):
            if x[-1] != 0:
                volOK += celfrac[1][i]

    vals = sum(valf)
    if not corr:
        err2 = sum(map(lambda x, y: x * x * y * y, valf, voxerrs))
        errs = np.divide(
            np.sqrt(err2), vals, out=np.zeros_like(err2), where=(vals != 0)
        )
    else:
        err = sum(map(lambda x, y: abs(x * y), valf, voxerrs))
        errs = np.divide(err, vals, out=np.zeros_like(err), where=(vals != 0))

    # Default units are X per volume units
    if Vmult == "vtot":
        # multiply by the total volume of the voxel

        vals = celfrac[0] * vals  # here units are integral X
    elif Vmult == "celf":
        # multiply by the voxel volume fraction
        vals = (
            vals / volOK
        )  # here units are X per volume units (volume of no zero voxel)

    return vals, errs


def _sumElements(voxvals, voxerrs, celfrac, Vsum="onef", Vmult="none",
                 corr=False):
    if Vsum == "onef":
        # sum all value in the voxel
        vals = voxvals[0]
        errs = voxerrs[0]

        # insert value and error total over all energy bins
        xs = sum(vals)
        if xs != 0:
            ve = vals * errs
            if not corr:
                err2 = sum(ve * ve)
                errt = np.sqrt(err2) / xs
            else:
                errt = sum(abs(ve)) / xs
            errs = np.append(voxerrs[0], errt)
        else:
            errs = np.append(voxerrs[0], 0)
        vals = np.append(vals, xs)
    else:
        # sum the value in the voxel multiplied by the volume fraction
        if len(voxvals) > 1:
            volOK = 0
            valf = map(lambda x, y: x * y, voxvals, celfrac[1])

            # insert value and error total over all energy bins
            for i, x in enumerate(valf):
                xs = sum(x)
                if xs != 0:
                    volOK += celfrac[1][i]
                    ve = x * voxerrs[i]
                    if not corr:
                        err2 = sum(ve * ve)
                        errt = np.sqrt(err2) / xs
                    else:
                        errt = sum(abs(ve)) / xs
                    voxerrs[i] = np.append(voxerrs[i], errt)
                else:
                    voxerrs[i] = np.append(voxerrs[i], 0)
                valf[i] = np.append(x, xs)

            vals = sum(valf)

            if not corr:
                err2 = sum(map(lambda x, y: x * x * y * y, valf, voxerrs))
                errs = np.divide(
                    np.sqrt(err2), vals, out=np.zeros_like(err2),
                    where=(vals != 0)
                )
            else:
                err = sum(map(lambda x, y: abs(x * y), valf, voxerrs))
                errs = np.divide(err, vals, out=np.zeros_like(err),
                                 where=(vals != 0))
        else:
            volOK = celfrac[1][0]
            vals = voxvals[0]
            errs = voxerrs[0]

            # insert value and error total over all energy bins
            xs = sum(vals)
            if xs != 0:
                ve = vals * errs
                if not corr:
                    err2 = sum(ve * ve)
                    errt = np.sqrt(err2) / xs
                else:
                    errt = sum(abs(ve)) / xs
                errs = np.append(voxerrs[0], errt)
            else:
                errs = np.append(voxerrs[0], 0)
            vals = np.append(vals, xs)

    if Vmult == "vtot":
        # multiply by the total volume of the voxel
        vals = celfrac[0] * vals
    elif Vmult == "celf":
        # multiply by the voxel volume fraction
        vals = vals / volOK

    return vals, errs


def identical_mesh(m1, m2):
    part, mesh, mtype = True, True, True
    if m1.part != m2.part:
        part = False

    if not m1.sameMesh(m2):
        mesh = False

    if m1.type != m2.type:
        mtype = False

    return part, mesh, mtype


# def scalemesh(m1, f1):
#     mscal = Fmesh(m1.meshtal)
#     mscal.filled = True
#     mscal.dims = m1.dims
#     mscal.ldims = m1.ldims
#     mscal.cart = m1.cart
#     mscal.part = m1.part
#     mscal.etag = m1.etag
#     mscal.type = m1.type
#     mscal.origin = m1.origin
#     mscal.rotation = m1.rotation
#     if not m1.cart:
#         mscal.axis = m1.axis
#         mscal.vec = m1.vec

#     mscal.dat = m1.dat * f1
#     mscal.err = m1.err
#     return mscal


# # sum two meshes with the same dimemsions
# def addmesh(m1, m2, f1=1.0, f2=1.0, corr=False):
#     if m1.part != m2.part:
#         print("Warning: particle type are different")

#     if not m1.sameMesh(m2):
#         print("mesh dimensions are not equal")
#         return None

#     if m1.type != m2.type:
#         print("mesh type are different")
#         return None

#     msum = Fmesh(m1.meshtal)

#     msum.filled = True
#     msum.dims = m1.dims[:]
#     msum.ldims = m1.ldims[:]
#     msum.cart = m1.cart
#     msum.part = m1.part
#     msum.etag = m1.etag
#     msum.type = m1.type
#     msum.origin = m1.origin
#     msum.rotation = m1.rotation
#     if not m1.cart:
#         msum.axis = m1.axis
#         msum.vec = m1.vec

#     msum.dat = m1.dat * f1 + m2.dat * f2

#     if not corr:
#         sf1 = f1 * f1
#         sf2 = f2 * f2
#         numerator = np.sqrt(sf1 * (m1.dat * m1.err) ** 2 + sf2 * (m2.dat * m2.err) ** 2)
#         msum.err = np.divide(
#             numerator, msum.dat, out=np.zeros_like(numerator), where=(msum.dat != 0)
#         )
#     else:
#         numerator = abs(f1 * m1.dat * m1.err) + abs(f2 * m2.dat * m2.err)
#         msum.err = np.divide(
#             numerator, msum.dat, out=np.zeros_like(numerator), where=(msum.dat != 0)
#         )

#     return msum


# # provide difference between two meshes
# def diffmesh(m1, m2, absvalue=False, relative=False):
#     if m1.part != m2.part:
#         print("Warning: particle type are different")

#     if not m1.sameMesh(m2):
#         print("mesh dimensions are not equal")
#         return None

#     if m1.type != m2.type:
#         print("mesh type are different")
#         return None

#     msum = Fmesh(m1.meshtal)

#     msum.filled = True
#     msum.dims = m1.dims[:]
#     msum.ldims = m1.ldims[:]
#     msum.cart = m1.cart
#     msum.part = m1.part
#     msum.etag = m1.etag
#     msum.type = m1.type
#     msum.origin = m1.origin
#     msum.rotation = m1.rotation
#     if not m1.cart:
#         msum.axis = m1.axis
#         msum.vec = m1.vec

#     if absvalue:
#         msum.dat = abs(m1.dat) - abs(m2.dat)
#     else:
#         msum.dat = m1.dat - m2.dat

#     denominator = np.sqrt((m1.dat * m1.err) ** 2 + (m2.dat * m2.err) ** 2)
#     msum.err = np.divide(
#         msum.dat, denominator, out=np.zeros_like(msum.dat),
#         where=(denominator != 0)
#     )

#     if relative:
#         denominator = m1.dat + m2.dat
#         msum.err = np.divide(
#             2 * msum.dat,
#             denominator,
#             out=np.zeros_like(msum.dat),
#             where=(denominator != 0),
#         )

#     return msum


# # sum two meshes with the same dimemsions
# def addbin(m1, binlist, flist=[], corr=False):
#     nbins = len(binlist)
#     if flist != []:
#         if len(flist) != nbins:
#             print("bin length and factor lenght is different")
#             return None

#     if m1.etag == "times":
#         print("Times bin cannot be added")
#         return None

#     if nbins >= m1.ldims[0]:
#         print("too many bin numbers")
#         return None

#     for b in binlist:
#         if b >= m1.ldims[0] - 1:
#             print("at least one bin index exceed mesh bin length")
#             return None

#     msum = Fmesh(m1.meshtal)
#     msum.dtype = m1.dtype

#     msum.filled = True
#     msum.dims = m1.dims[:]
#     msum.ldims = m1.ldims[:]
#     msum.part = m1.part
#     msum.etag = m1.etag
#     msum.type = m1.type
#     msum.origin = m1.origin
#     msum.rotation = m1.rotation
#     if not m1.cart:
#         msum.axis = m1.axis
#         msum.vec = m1.vec
#     msum.cart = m1.cart

#     msum.ldims[0] = 1
#     msum.dims[0] = np.array([0], m1.dtype)

#     msum.dat = np.zeros(msum.ldims, m1.dtype)
#     msum.err = np.zeros(msum.ldims, m1.dtype)

#     if flist == []:
#         for b in binlist:
#             msum.dat[0, :, :, :] += m1.dat[b, :, :, :]
#             if not corr:
#                 msum.err[0, :, :, :] += (m1.err[b, :, :, :] * m1.dat[b, :, :, :]) ** 2
#             else:
#                 msum.err[0, :, :, :] += abs(m1.err[b, :, :, :] * m1.dat[b, :, :, :])
#     else:
#         for i, b in enumerate(binlist):
#             msum.dat[0, :, :, :] += m1.dat[b, :, :, :] * flist[i]
#             if not corr:
#                 msum.err[0, :, :, :] += (
#                     m1.err[b, :, :, :] * m1.dat[b, :, :, :] * flist[i]
#                 ) ** 2
#             else:
#                 msum.err[0, :, :, :] += abs(
#                     m1.err[b, :, :, :] * m1.dat[b, :, :, :] * flist[i]
#                 )

#     if not corr:
#         msum.err = np.divide(
#             np.sqrt(msum.err),
#             msum.dat,
#             out=np.zeros_like(msum.err),
#             where=(msum.dat != 0),
#         )
#     else:
#         msum.err = np.divide(
#             msum.err, msum.dat, out=np.zeros_like(msum.err),
#             where=(msum.dat != 0)
#         )

#     return msum


# ================== END OF FUNCTION DEFINITIONS =================



"""
This module is related to the parsing and manipulation of MCNP input files.

The parser is built on the numjuggler python module.

"""
from __future__ import annotations
"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/ 
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import os
import logging
import json
import re
from numjuggler import parser
import pandas as pd

from f4enix.input.materials import MatCardsList, Material
from f4enix.input.libmanager import LibManager
from f4enix.input.auxiliary import debug_file_unicode
from f4enix.constants import PAT_COMMENT, PAT_CARD_KEY, PAT_FMESH_KEY, PAT_NP
from f4enix.input.d1suned import ReactionFile, IrradiationFile, Reaction
from copy import deepcopy


PAT_MT = re.compile(r'm[tx]\d+', re.IGNORECASE)
PAT_BLANK_LINE = re.compile(r'\n[\s\t]*\n')
ADD_LINE_FORMAT = '         {}\n'


class Input:
    def __init__(
            self,
            cells: list[parser.Card],
            surfs: list[parser.Card],
            data: list[parser.Card],
            header: list = None) -> None:
        """Class representing an MCNP input file.

        cells, surfaces, materials and transformations are handled explicitly.
        All other datacards are treated generically for the moment being.
        The parsing is built on the numjuggler python module.

        Parameters
        ----------
        cells : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP cells.
        surfs : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP surfaces.
        data : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP data cards.
        header : list, optional
            list of strings that compose the MCNP header, by default None

        Attributes
        ----------
        cells: dict[str, parser.Card]
            cleaned numjuggler cards for each cells in the input. keys are the
            number of the cells.
        surfs: dict[str, parser.Card]
            cleaned numjuggler cards for each surface in the input. keys are
            the number of the surfaces.
        materials: MatCardsList
            material cards section of the input.
        transformations: list[parser.Card]
            list of transformation datacards (i.e. TRn)
        other_data: list[parser.Card]
            list of all remaining datacards that are treated in a generic way
        tally_keys: list[int]
            ids of the tallies available in the input
        fmesh_keys: list[int]
            ids of the fmeshes available in the input

        Examples
        --------
        The most common way to initiliaze an Input object is from a MCNP input
        file

        >>> from f4enix.input.MCNPinput import Input
        ... # Read the input file
        ... inp = Input.from_input(inpfile)

        >>> inp.cells
        ... # inp.surfs
        {'1': <numjuggler.parser.Card at 0x26535f02a70>,
         '2': <numjuggler.parser.Card at 0x26535f03130>,
         ...
         '128': <numjuggler.parser.Card at 0x26535f48f70>}

        The input can be translated to another library and rewritten to a file

        >>> from f4enix.input.libmanager import LibManager
        ... # Initialize a default nuclear data libraries manager
        ... libmanager = LibManager()
        ... # Translate the input to another library
        ... inp.translate('21c', libmanager)
        ... inp.write(outfile_path)

        Retrieve (and possibly modify) different cards in the input

        >>> print('Cell number 1:')
        ... print(inp.get_cells_by_id([1]))
        ... print('Surfaces number 10 and 20:')
        ... print(inp.get_surfs_by_id([10, 20]))
        ... print('All cells with material M1')
        ... print(inp.get_cells_by_matID(1))
        ... # use this method only if the previous ones
        ... # are not enough
        ... print('Generic way to obtain cards')
        ... print(inp._get_cards_by_id(['SDEF', 'IMP:N,P'], inp.other_data))
        Cell number 1:
        {'1': <numjuggler.parser.Card object at 0x0000026535F02A70>}
        Surfaces number 10 and 20:
        {'10': <numjuggler.parser.Card object at 0x0000026535F491B0>,
        '20': <numjuggler.parser.Card object at 0x0000026535F49390>}
        All cells with material M1
        {'22': <numjuggler.parser.Card object at 0x00000265360E7850>}
        Generic way to obtain cards
        {'SDEF': <numjuggler.parser.Card object at 0x0000026535F4B190>,
        'IMP:N,P': <numjuggler.parser.Card object at 0x0000026535F4A890>}

        Extract a subset of cells depending on some condition and from that
        dump out a minimal working MCNP input that includes all necessary
        surfaces, transformations and materials.

        >>> # --- Extract cells based on material ---
        ... inp = Input.from_input(inpfile)
        ... cells_ids = []  # store here the cells to be extracted
        ... selected_mat = 11
        ... for key, cell in inp.cells.items():
        ...     # get the material of the cell using numjuggler API
        ...     mat = cell._get_value_by_type('mat')
        ...     if mat == selected_mat:
        ...         cells_ids.append(key)
        ... print(cells_ids)
        ... # extract the cells subset to a file
        ... inp.extract_cells(cells_ids, 'outfile.i')

        Get useful summary of the material section of the input in a
        pandas.DataFrame object.

        >>> from f4enix.input.MCNPinput import Input
        ... from f4enix.input.libmanager import LibManager
        ... # Initialize a default nuclear data libraries manager
        ... libmanager = LibManager(xsdir_file='my_xsdir')
        ... inp = Input.from_input(inpfile)
        ... inp.materials.get_info(libmanager, complete=True)
        (                              Atom Fraction  Mass Fraction
        Material Submaterial Element
        m1       1           H             0.021630      -0.019046
                 2           C             0.018920      -0.198524
                 3           N             0.002060      -0.025207
                 4           O             0.027060      -0.378226
                 5           Mg            0.001190      -0.025268
        ...                                     ...            ...
        M29      1           Nb            0.000005      -0.000100
                             Co            0.000041      -0.000500
                             Fe            0.055451      -0.648443
        M30      1           H             0.063398      -0.112140
                             O             0.031622      -0.887860
        [333 rows x 2 columns],
                                    Fraction  Sub-Material Fraction  \
        Material Submaterial Element
        M24      1           Ag       0.000016               0.000016
                             Al       0.000631               0.000631
                             As       0.000023               0.000023
                             Au       0.000009               0.000009
                             Bi       0.000163               0.000163
        ...                                ...                    ...
        m7       2           O        0.033428               1.000000
        m8       1           Be       0.002970               0.034219
        ...
                             Cu                0.944769
                             Ni                0.021012
        m9       1           Be                1.000000
        [333 rows x 3 columns])


        """

        self.cells = self._to_dict(cells)
        self.surfs = self._to_dict(surfs)

        (self.materials, self.transformations,
         self.other_data) = self._parse_data_section(data)

        self.header = header

        # get a list of the tally keys
        tally_keys = []
        fmesh_keys = []
        for key, card in self.other_data.items():
            if card.dtype == 'Fn':
                tally_keys.append(card.name)
            elif PAT_FMESH_KEY.match(key):
                fmesh_keys.append(card.name)
        self.tally_keys = tally_keys
        self.fmesh_keys = fmesh_keys

        # # store also all the original cards for compatibility
        # # with some numjuggler modes
        # cells.extend(surfs)
        # cells.extend(data)

    @classmethod
    def from_input(cls, inputfile: os.PathLike) -> Input:
        """ Generate an Input object from an MCNP input text file using
        numjuggler as a parser

        Parameters
        ----------
        inputfile : os.PathLike
            input file

        Returns
        -------
        Input
            Input object
        """
        cells, surfaces, data, header = _get_input_arguments(inputfile)

        return cls(cells, surfaces, data, header=header)

    def write(self, outfilepath: os.PathLike, wrap: bool = False) -> None:
        """write the input to a file

        Parameters
        ----------
        outfilepath : os.PathLike
            path to the output file
        wrap : bool
            if true the text is wrapped at 80 char. May cause slowdowns
        """
        logging.info('Writing to {}'.format(outfilepath))

        with open(outfilepath, 'w') as outfile:

            # Add the header lines
            for line in self.header:
                outfile.write(line)
            # Add the cells
            outfile.writelines(self._print_cards(self.cells, wrap=wrap))
            # Add a break
            outfile.write('\n')
            # Add the surfaces
            outfile.writelines(self._print_cards(self.surfs, wrap=wrap))
            # Add a break
            outfile.write('\n')
            # Add the material section (they exit without the \n)
            outfile.write(self.materials.to_text()+'\n')
            # Add the translations
            outfile.writelines(self._print_cards(self.transformations,
                                                 wrap=wrap))
            # Add the rest of the datacards
            outfile.writelines(self._print_cards(self.other_data, wrap=wrap))
            # Add a break
            outfile.write('\n')

        logging.info('File was written correctly')

    def translate(self, newlib: str, libmanager: LibManager) -> None:
        """
        Translate the input to another library

        Parameters
        ----------
        newlib : dict | str
            There are a few ways that newlib can be provided:

            1) str (e.g. 31c), the new library to translate to will be the
            one indicated;

            2) dic (e.g. {'98c' : '99c', '31c: 32c'}), the new library is
            determined based on the old library of the zaid

            3) dic (e.g. {'98c': [list of zaids], '31c': [list of zaids]}),
            the new library to be used is explicitly stated depending
            on the zaidnum.

        libmanager : libmanager.LibManager
            Library manager for the conversion.

        Returns
        -------
        None.

        """

        try:
            if newlib[0] == '{':
                # covert the dic
                newlib = json.loads(newlib)
        except KeyError:
            # It is already a dict, pass
            pass

        self.update_zaidinfo(libmanager)
        self.materials.translate(newlib, libmanager)

    def update_zaidinfo(self, lib_manager: LibManager):
        """
        This methods allows to update the in-line comments for every zaids
        containing additional information

        Parameters
        ----------
        lib_manager : libmanager.LibManager
            Library manager for the conversion.

        Returns
        -------
        None.

        """

        self.materials.update_info(lib_manager)

    @staticmethod
    def set_cell_void(cell: parser.Card) -> None:
        if cell.ctype == 3 and cell.get_m() != 0:
            cell.hidden['~'][0] = ''
            cell._set_value_by_type('mat', 0)
            cell._Card__m = 0  # necessary for the get val
        else:
            logging.warning(
                f'cell {cell.name} is either already void or not a cell')

    @staticmethod
    def _print_cards(cards: dict[str, parser.Card],
                     wrap: bool = False) -> list[str]:
        text = []
        for _, card in cards.items():
            text_candidate = card.card(wrap=wrap).strip('\n')+'\n'
            # avoid blank lines
            text_candidate = PAT_BLANK_LINE.sub('\n', text_candidate)
            text.append(text_candidate)
        return text

    @staticmethod
    def _to_dict(cards: list[parser.Card]) -> dict[str, parser.Card]:
        new_cards = {}
        flag_add = False
        for card in cards:

            card.get_values()
            try:
                key = card.name
                key = card.card().split()[0].upper()
                if key in new_cards.keys():
                    raise KeyError('Duplicated card entry: '+key)

            except AttributeError:
                # This means that this is a fake card just made by comments
                # it should be merged with the following card

                # this is true for comments, but sometimes it happens also
                # with real cards due to bugs in numjuggler

                # let's first check if it is a comment
                if PAT_COMMENT.match(card.lines[0]) is not None:
                    comment = card.lines
                    flag_add = True
                    continue
                # and then if it is a proper card
                else:
                    key = card.card().split()[0].upper()
                    if key in new_cards.keys():
                        raise KeyError('Duplicated card entry: '+key)

            if flag_add:
                comment.extend(card.lines)
                card.lines = comment
                card.get_input()
                card.get_values()
                flag_add = False

            new_cards[key] = card

        return new_cards

    @staticmethod
    def _get_cards_by_id(ids: list[str], cards: dict
                         ) -> dict[str, parser.Card]:
        selected_cards = {}
        for id_card in ids:
            try:
                selected_cards[id_card] = cards[id_card]
            except KeyError:
                try:
                    # sometimes it may be with an asterisk
                    selected_cards['*'+id_card] = cards['*'+id_card]
                except KeyError:
                    raise KeyError('The card is not available in the input')

        return selected_cards

    def get_cells_by_id(self, ids: list[int]) -> dict[str, parser.Card]:
        """given a list of cells id return a dictionary of such cells

        Parameters
        ----------
        ids : list[int]
            cells id to be extracted

        Returns
        -------
        dict
            extracted cells
        """
        str_ids = []
        for id in ids:
            str_ids.append(str(id))
        return self._get_cards_by_id(str_ids, self.cells)

    def get_surfs_by_id(self, ids: list[int]) -> dict[str, parser.Card]:
        """given a list of surfaces id return a dictionary of such surfaces

        Parameters
        ----------
        ids : list[int]
            cells id to be extracted

        Returns
        -------
        dict
            extracted surfaces
        """
        str_ids = []
        for id in ids:
            str_ids.append(str(id))
        return self._get_cards_by_id(str_ids, self.surfs)

    def get_materials_subset(self, ids: list[str] | str) -> MatCardsList:
        """given a list of material ids generate a new MatCardsList with
        the requested subset

        Parameters
        ----------
        ids : Union(list[str]), str)
            ids of the materials to put into the subset

        Returns
        -------
        MatCardsList
            new materials subset
        """
        if type(ids) is str:
            return self.materials[ids.upper()]
        else:
            materials = []
            for id_mat in ids:
                materials.append(self.materials[id_mat.upper()])
            return MatCardsList(materials)

    def get_data_cards(self, ids: list[str] | str) -> dict[str, parser.Card]:
        """Get a tranformation card or an other data card by its key.

        For the moment, transformation cards mixed with other data cards is
        not supported.

        Parameters
        ----------
        ids : list[str] | str
            keys of the cards to retrieve

        Returns
        -------
        dict[str, parser.Card]
            retrieved cards
        """
        if type(ids) is str:
            ids = [ids]

        try:
            cards = self._get_cards_by_id(ids, self.other_data)
        except KeyError:
            cards = self._get_cards_by_id(ids, self.transformations)

        return cards

    def _parse_data_section(self, cards: list[parser.Card]
                            ) -> tuple[MatCardsList,
                                       list[parser.Card],
                                       list[parser.Card]]:

        # first of all correct numjuggler parser
        cards = self._to_dict(cards)

        materials = []  # store here the materials
        transformations = {}  # store translations
        other_data = {}  # store here the other data cards

        for key, card in cards.items():
            key = self._clean_card_name(key)
            try:
                if card.values[0][1] == 'mat':
                    if (PAT_MT.match(card.lines[0]) or
                            PAT_MT.match(card.lines[-1])):
                        # mt or mx cards should be added to the previous
                        # material
                        materials[-1].add_mx(card)
                    else:
                        materials.append(Material.from_text(card.lines))

                elif card.dtype == 'TRn':
                    transformations[key] = card

                else:
                    other_data[key] = card

            except IndexError:
                # this means that there were no values
                other_data[key] = card

        return MatCardsList(materials), transformations, other_data

    def extract_cells(self, cells: list[int], outfile: os.PathLike,
                      renumber_from: int = None, keep_universe:bool = True):
        """given a list of cells, dumps a minimum MCNP working file that
        includes all the requested cells, defined surfaces, materials and
        translations.
        TODO: if keep_universe is False: the cell definitions of the original input
         class will be modified. This may be undesirable.

        Parameters
        ----------
        cells : list[int]
            desired list of cells
        outfile : os.PathLike
            path to the file where the MCNP input needs to be dumped
        renumber_from : int
            apply a renumbering to the extracted cells starting from the 
            specified int. It is important to notice that this renumbering
            DOES NOT SUPPORT # operator for the moment being.
            Default is None, no renumbering is applied.
        """
        logging.info('Collecting the cells, surfaces, materials and transf.')
        # make sure these are str
        cset = []
        for cell in cells:
            cset.append(str(cell))
        cset = set(cset)

        # first, get all surfaces needed to represent the cn cell.
        sset = set()  # surfaces
        mset = set()  # material
        # tset = set()  # transformations

        # next runs: find all other cells:
        again = True
        while again:
            again = False
            for key, c in self.cells.items():
                if key in cset:
                    cref = c.get_refcells()
                    if cref.difference(cset):
                        again = True
                        cset = cset.union(cref)

        # Get all surfaces and materials
        cells_cards = self.get_cells_by_id(cset)
        for i, (_, cell) in enumerate(cells_cards.items()):
            for v, t in cell.values:
                if t == 'sur':
                    sset.add(v)
                elif t == 'mat':
                    if int(v) != 0:  # void material is not defined in a card
                        mset.add('M'+str(v))
            if renumber_from is not None:
                cell._set_value_by_type('cel', i+renumber_from)
            if not keep_universe:
                new_input = []
                for input_part in cell.input:
                    new_input.append(re.sub(r"[uU]=\{:<\d+\}", "", input_part))    
                cell.input = new_input

        # Do not bother for the moment in selecting also the transformations

        #                     elif t == 'tr':
        #                         tset.add(v)

        # # final run: for all cells find surfaces, materials, etc.
        # for key, surf in self.surfs.items():
        #     if key in sset:
        #         # surface card can refer to tr
        #         for v, t in surf.values:
        #             if t == 'tr':
        #                 tset.add(v)

        logging.info('write MCNP reduced input')
        with open(outfile, 'w') as outfile:
            # Add the header lines
            for line in self.header:
                outfile.write(line)
            # Add the cells
            outfile.writelines(self._print_cards(cells_cards))
            # Add a break
            outfile.write('\n')
            # Add the surfaces
            surfs = self.get_surfs_by_id(sset)
            outfile.writelines(self._print_cards(surfs))
            # Add a break
            outfile.write('\n')
            # Add materials
            materials = self.get_materials_subset(mset)
            outfile.write(materials.to_text()+'\n')
            outfile.writelines(self._print_cards(self.transformations))

        logging.info('input written correctly')

    def extract_universe(self, universe: int, outfile: os.PathLike):
        """Dumps a minimum MCNP working file that
        includes all the cells, surfaces, materials and
        translations of the universe. The resulting file doesn't have the universe
        keyword in the cell definitions

        Parameters
        ----------
        universe : int
            universe id to be extracted
        outfile : os.PathLike
            path to the file where the MCNP input needs to be dumped
        """
        cell_ids_to_extract = []
        for cell_id, cell in self.cells.items():
            cell_universe = cell.get_u()

            if cell_universe == universe:
                cell_ids_to_extract.append(cell_id)
                
        self.extract_cells(
            cells=cell_ids_to_extract, 
            outfile=outfile, 
            keep_universe=False
            )

    @staticmethod
    def _clean_card_name(key: str) -> str:
        # this is to clean cases like:
        # *TR1 -> TR1
        # F6:N,P -> F6
        try:
            newkey = PAT_CARD_KEY.search(key).group()
        except AttributeError:
            logging.debug('the following key was not cleaned: '+key)
            newkey = key

        return newkey

    def get_cells_by_matID(self, matID: int,
                           deepcopy_flag: bool = True
                           ) -> dict[str, parser.Card]:
        """Given a material ID return a dictionary {key, card} of all
        the cells to which that material is assigned to.

        The cells that are returned are deepcopies of the original ones.

        Parameters
        ----------
        matID : int
            material ID to filter the cells
        deepcopy_flag: bool
            if False, the cells are not copied. Default is True

        Returns
        -------
        dict[int, parser.Card]
            cells to which the material is assigned to
        """
        logging.debug('get cells for material {} requested'.format(matID))
        filtered_cells = {}
        for key, cell in self.cells.items():
            if cell._get_value_by_type('mat') == int(matID):
                if deepcopy_flag:
                    filtered_cells[key] = deepcopy(cell)
                else:
                    filtered_cells[key] = cell
        return filtered_cells

    def scale_densities(self, factor: float) -> None:
        """Scale the density values of all cells by the same factor. Void
        cells are ignored. Resulting density will be equal to
        original_density*factor

        Parameters
        ----------
        factor : float
            scaling factors for the densities
        """
        for key, cell in self.cells.items():
            if not cell._get_value_by_type('mat') == 0:
                density = cell.get_d()
                newdensity = float('{:.5e}'.format(density*factor))
                cell.set_d(newdensity)

    def get_cells_summary(self) -> pd.DataFrame:
        """Get a summary of infos for each cell

        A DataFrame is returned where for each cell is listed the material,
        density, universe and filler is present.

        Returns
        -------
        pd.DataFrame
            Summary of cells info
        """
        rows = []
        for key, cell in self.cells.items():
            row = {'cell': int(key)}
            row['material'] = cell.get_m()
            row['density'] = cell.get_d()
            row['universe'] = cell.get_u()
            row['filler'] = cell.get_f()
            rows.append(row)

        df = pd.DataFrame(rows)
        return df.set_index('cell').sort_index()

    def _get_tally_cards(self, idx: int,
                         ) -> list[str]:
        keys = []
        pat = re.compile(r'F[a-zA-Z]*{}$'.format(idx))
        for key, _ in self.other_data.items():
            if pat.match(key) is not None:
                keys.append(key)
        return keys

    def _retrieve_input(self, tag: str) -> str:
        # get the FC comment excluding the FC tag
        comment_line = self.other_data[tag].input[0]
        inp = comment_line.replace(tag, '').strip()
        inp = inp.replace(tag.lower(), '').strip()
        return inp

    def get_tally_summary(self, fmesh: bool = False) -> pd.DataFrame:
        """Get a summary of the tallies defined in the input

        Both normal tallies and fmeshes can be requested. For each tally the
        number, particle, description and multiplier are listed (if available)

        Parameters
        ----------
        fmesh : bool, optional
            if True produced a summary for the fmehses instead of for the
            normal tallies, by default False

        Returns
        -------
        pd.DataFrame
            summary info on defined tallies

        """

        if fmesh:
            tag_tally = 'FMESH'
            tallies = self.fmesh_keys
        else:
            tag_tally = 'F'
            tallies = self.tally_keys

        rows = []
        for key in tallies:
            desc = pd.NA
            particle = pd.NA
            multiplier = None
            card_keys = self._get_tally_cards(key)
            for aux_key in card_keys:
                if aux_key[:2] == 'FC':
                    desc = self._retrieve_input(aux_key)
                elif aux_key == tag_tally+str(key):
                    line = self.other_data[aux_key].input[0]
                    particle = PAT_NP.search(line).group().upper().strip(':')
                elif aux_key[:2] == 'FM':
                    multiplier = self._retrieve_input(aux_key).split()

            row = {'Tally': key, 'Particle': particle, 'Description': desc}

            if multiplier is not None:
                row['Normalization'] = multiplier[0]
                if len(multiplier) > 1:
                    row['Other multipliers'] = multiplier[1:]
                else:
                    row['Other multipliers'] = pd.NA
            else:
                row['Normalization'] = pd.NA
                row['Other multipliers'] = pd.NA

            rows.append(row)

        return pd.DataFrame(rows).set_index("Tally").sort_index()

    def replace_material(self, new_mat_id: int,
                         new_density: str,
                         old_mat_id: int,
                         u_list: list[int] = None
                         ) -> None:
        """Replace a material and density in the input with other values.

        Parameters
        ----------
        new_mat_id : int
            id of the new material (0 for void)
        new_density : str
            new value for the density (including sign)
        old_mat_id : int
            id of the material to be repèlaced
        u_list : list[int]
            change the material only if cells belong to one of the universes
            in the list. By default is None, all cells are affected.

        Raises
        ------
        NotImplementedError
            The capability to switch from a void cell to a filled cell is not
            implemented yet. Viceversa is possible.
        """
        if old_mat_id == 0:
            raise NotImplementedError('Cannot change a void cell')

        for _, cell in self.cells.items():
            in_universe = False
            # check if universe is a parameter
            if u_list is not None:
                if cell.get_u() in u_list:
                    in_universe = True
            else:
                # always in universe, default is always
                in_universe = True

            # If the material needs change and in universe
            if cell.get_m() == old_mat_id and in_universe:
                # Void needs to be handle in a specific way
                if new_mat_id == 0:
                    Input.set_cell_void(cell)
                else:
                    cell._set_value_by_type('mat', new_mat_id)
                    cell._Card__m = new_mat_id  # necessary for the get val
                    cell.set_d(new_density)


class D1S_Input(Input):

    def __init__(self, cells: list[parser.Card], surfs: list[parser.Card],
                 data: list[parser.Card], header: list = None,
                 irrad_file: IrradiationFile = None,
                 reac_file: ReactionFile = None) -> None:
        """Children of the :py:class:`Input`.

        it includes also the reaction and irradiation files necessary for a 
        D1S-UNED run and defines additional methods related to them.

        Parameters
        ----------
        cells : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP cells.
        surfs : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP surfaces.
        data : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP data cards.
        header : list, optional
            list of strings that compose the MCNP header, by default None
        irrad_file : IrradiationFile, optional
            irradiation file object, by default None
        reac_file : ReactionFile, optional
            readtion file object, by default None

        Attributes
        ----------
        irrad_file : IrradiationFile
            irradiation file object
        reac_file : ReactionFile
            readtion file object

        Examples
        --------
        translate the input defining an activation and transport library.
        the reaction file will be used to identify to which isotopes the
        activation library has to be assigned.

        >>> from f4enix.input.MCNPinput import D1S_Input
        ... from f4enix.input.libmanager import LibManager
        ... d1s_inp = D1S_Input.from_input('d1stest.i', irrad_file='irr_test',
        ...                                reac_file='reac_fe')
        ... d1s_inp.smart_translate('99c', '00c', LibManager())
        """

        super().__init__(cells, surfs, data, header=header)
        self.irrad_file = irrad_file
        self.reac_file = reac_file

    @classmethod
    def from_input(cls, inputfile: os.PathLike, irrad_file: os.PathLike = None,
                   reac_file: os.PathLike = None) -> D1S_Input:
        """Generate a D1S-UNED input file.

        this includes also the reaction and irradiation files.

        Parameters
        ----------
        inputfile : os.PathLike
            path to the MCNP input (D1S)
        irrad_file : os.PathLike, optional
            path to the irradiation file, by default None (no file associated)
        reac_file : os.PathLike, optional
            path to the reaction file, by default None (no file associated)

        Returns
        -------
        D1S_Input
            generated D1S_Input object
        """
        cells, surfaces, data, header = _get_input_arguments(inputfile)
        if irrad_file is not None:
            irrad_file = IrradiationFile.from_text(irrad_file)
        if reac_file is not None:
            reac_file = ReactionFile.from_text(reac_file)

        return D1S_Input(cells, surfaces, data, header=header,
                         irrad_file=irrad_file, reac_file=reac_file)

    def get_potential_paths(self, libmanager: LibManager,
                            lib: str) -> list[Reaction]:
        """Given an activation library, return a list of all possible reactions
        paths foreseen by the libmanager that can originate from the material
        section of the input.

        Parameters
        ----------
        libmanager : LibManager
            Handlers of cross sections operations
        lib : str
            activation library to be used (e.g. 99c)

        Returns
        -------
        list[Reaction]
            list of Reaction objects describing all possible paths included in
            LibManager
        """
        reactions = []
        for material in self.materials.materials:
            for submat in material.submaterials:
                for zaid in submat.zaidList:
                    parent = zaid.element+zaid.isotope
                    zaidreactions = libmanager.get_reactions(lib, parent)
                    # if len(zaidreactions) > 0:
                    #     # it is a parent only if reactions are available
                    #     parentlist.append(parent)
                    for MT, daughter in zaidreactions:
                        reactions.append((parent, MT, daughter))
                        # daughterlist.append(daughter)

        reactions = list(set(reactions))
        reactions.sort()
        # --- Build the reactions and reaction file ---
        reaction_list = []
        for parent, MT, daughter in reactions:
            parent = parent+'.'+lib
            # Build a comment
            _, parent_formula = libmanager.get_zaidname(parent)
            _, daughter_formula = libmanager.get_zaidname(daughter)
            comment = '{} -> {}'.format(parent_formula, daughter_formula)

            rx = Reaction(parent, MT, daughter, comment=comment)
            reaction_list.append(rx)

        return reaction_list

    def get_reaction_file(self, libmanager: LibManager, lib: str,
                          set_as_attribute: bool = True
                          ) -> ReactionFile:
        """
        Get a reaction file suitable for the input.

        The reaction file is built selecting from all the possible reaction
        paths that can originate in the model due to its material cards only
        the reactions that lead to a daughter listed in the irradiation file.
        By default this is added as the reac_file for the input.

        Parameters
        ----------
        libmanager : LibManager
            Object handling all cross-sections related operations.
        lib : str
            library suffix to be used.
        set_as_attribute: bool
            if True (default) the reactionfile is saved as the self.reac_file

        Returns
        -------
        ReactionFile
            Object representing the react file for D1S.

        Raises
        ------
        ValueError
            if no irradiation files have been assigned yet to the input

        """
        # irrad file is necessary for this operation
        # recover all available daughters
        if self.irrad_file is None:
            raise ValueError(
                'irrad_file attribute cannot be None for this operation')
        else:
            available_daughters = self.irrad_file.get_daughters()

        # Recover all possible reactions
        reactions = self.get_potential_paths(libmanager, lib)

        # perform the selection
        selected_reactions = []
        for reaction in reactions:
            if reaction.daughter in available_daughters:
                # add the reaction to the one to use
                selected_reactions.append(reaction)

        reac_file = ReactionFile(selected_reactions)

        if set_as_attribute:
            self.reac_file = reac_file

        return reac_file

    def smart_translate(self, activation_lib: str, transport_lib: str,
                        libmanager: LibManager,
                        fix_natural_zaid: bool = False) -> None:
        """
        Translate the input to another library without relying on old libs.

        Both the activation and transport libraries are changed. The reaction
        file and PIKMT cards are also translated.
        Differently from self.translate(),
        libraries are not modified based on old ones but based on the reaction
        file. This is, to all parent zaids listed in the reactions will be
        assigned the activation_lib, to all others the transport_lib.

        Parameters
        ----------
        activation_lib : str
            library to be used for activation, e.g., 99c
        activation_lib : dict[str, str]
            library to be used for activation, e.g., 31c
        libmanager : LibManager
            Library manager for the conversion.
        fix_natural_zaid: bool
            if True, and additional initial translation with the transport lib
            is done in order to expand the natural zaids. If the transport lib
            do not expand the natural zaid there may be issues.

        Returns
        -------
        None

        """
        if self.reac_file is None:
            raise ValueError(
                'reac_file cannot be None for this operation'
            )

        if fix_natural_zaid:
            # get a first translation to avoid issues with old natural zaids
            self.translate(transport_lib, libmanager)

        active_zaids = []
        transp_zaids = []

        for reaction in self.reac_file.reactions:
            # strip the lib from the parent
            parent = reaction.parent.split('.')[0]
            active_zaids.append(parent)
            reaction.change_lib(activation_lib)

        # Now check for the remaing materials in the input to be assigned
        # to transport
        for material in self.materials.materials:
            for submaterial in material.submaterials:
                for zaid in submaterial.zaidList:
                    zaidnum = zaid.element+zaid.isotope
                    if (zaidnum not in active_zaids and
                            zaidnum not in transp_zaids):
                        transp_zaids.append(zaidnum)

        newlib = {activation_lib: active_zaids, transport_lib: transp_zaids}

        # trigger translation of the PIKMT card
        self.add_PIKMT_card()

        # Translate the input with the new lib
        self.materials.translate(newlib, libmanager)

    def add_PIKMT_card(self) -> None:
        """
        Add a PIKMT card to the input file. If a PKMT card is already present
        this will be overridden.

        Returns
        -------
        None.

        """
        key = 'PIKMT'
        lines = [key+'\n']
        for parent in self.reac_file.get_parents():
            lines.append('         {}    {}\n'.format(parent, 0))

        card = parser.Card(lines, 5, -1)
        self.other_data[key] = card  # should override other PKMT cards

    def add_track_contribution(self, tallykey: str, zaids: list[str],
                               who: str = 'parent'):
        """
        Given a list of zaid add the FU bin in the requested tallies in order
        to collect the contribution of them to the tally.

        Parameters
        ----------
        tallykey : str
            ID of the tally onto which to operate (e.g. F4).
        zaids : list[str]
            list of zaid numbers of the parents/daughters (e.g. 1001).
        who : str, optional
            either 'parent' or 'daughter' specifies the types of zaids to
            be tracked. The default is 'parent'.

        Raises
        ------
        ValueError
            check for admissible who parameter.

        Returns
        -------
        bool
            return True if lines were added correctly

        """
        card = self.other_data[tallykey]
        num = str(_get_num_tally(tallykey))

        card.lines.append('FU'+num+' 0\n')

        if who == 'parent':
            for zaid in zaids:
                card.lines.append(ADD_LINE_FORMAT.format('-'+str(zaid)))
        elif who == 'daughter':
            for zaid in zaids:
                card.lines.append(ADD_LINE_FORMAT.format(zaid))
        else:
            raise ValueError(who+' is not an admissible "who" parameters')
        card.get_input()


def _get_input_arguments(inputfile: os.PathLike
                            ) -> tuple:
    name = os.path.basename(inputfile).split(".")[0]

    # Get the blocks using numjuggler parser
    logging.info('Reading file: {}'.format(name))
    jug_cards = parser.get_cards_from_input(inputfile)
    try:
        jug_cardsDic = parser.get_blocks(jug_cards)
    except UnicodeDecodeError as e:
        logging.error('The file contains unicode errors, scan initiated')
        txt = debug_file_unicode(inputfile)
        logging.error('The following error where encountered: \n'+txt)
        raise e

    logging.debug('Reading has finished')

    # Parse the different sections
    header = jug_cardsDic[2][0].lines
    cells = jug_cardsDic[3]
    surfaces = jug_cardsDic[4]
    data = jug_cardsDic[5]

    return cells, surfaces, data, header


def _get_num_tally(key: str) -> int:
    patnum = re.compile(r'\d+')
    try:
        num = patnum.search(key).group()
    except AttributeError:
        # The pattern was not found
        raise ValueError(key+' is not a valid tally ID')

    return int(num)



"""
This module is related to the parsing of a material section of a MCNP input

The information is organized as follows:

MatCardsList
     Material
         Submaterial
               Element
                   Zaid

"""

from __future__ import annotations

"""
Copyright 2019 F4E | European Joint Undertaking for ITER and the Development of
Fusion Energy (‘Fusion for Energy’). Licensed under the EUPL, Version 1.2 or - 
as soon they will be approved by the European Commission - subsequent versions
of the EUPL (the “Licence”). You may not use this work except in compliance
with the Licence. You may obtain a copy of the Licence at: 
    https://eupl.eu/1.2/en/   
Unless required by applicable law or agreed to in writing, software distributed
under the Licence is distributed on an “AS IS” basis, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the Licence permissions
and limitations under the Licence.
"""

import re
import pandas as pd
from numjuggler import parser as par
from collections.abc import Sequence
from decimal import Decimal
import copy
import sys
import os
from contextlib import contextmanager
from f4enix.constants import PAT_COMMENT, PAT_MAT, PAT_MX
from f4enix.input.libmanager import LibManager


# -------------------------------------
# == CLASSES FOR MATERIAL READING ==
# -------------------------------------
class Zaid:

    def __init__(self, fraction: str | float, element: str, isotope: str,
                 library: str, ab: str = '', fullname: str = '') -> None:
        """
        Object representing a Zaid

        Parameters
        ----------
        fraction : str/float
            fraction of the zaid.
        element : str
            element part of the zaid (AA).
        isotope : str
            isotope part of the zaid (ZZZ).
        library : str
            library suffix (e.g. 99c).
        ab : str, optional
            abundance of the zaid in the material. The default is ''.
        fullname : str, optional
            formula name (e.g. H1). The default is ''.

        Returns
        -------
        None.

        All the init parameters are saved as attributes. In addition:

        Attributes
        ----------
        name: str
            AAZZZ[lib]

        """
        self.fraction = float(fraction)
        self.element = element
        self.isotope = isotope
        self.library = library
        self.ab = ab
        self.fullname = fullname

        if self.library is None:
            self.name = self.element+self.isotope
        else:
            self.name = self.element+self.isotope+'.'+self.library

    @classmethod
    def from_string(cls, string: str) -> Zaid:
        """
        Generate a zaid object from an MCNP string

        Parameters
        ----------
        string : str
            original MCNP string.

        Returns
        -------
        Zaid
            created zaid.

        """
        # Divide fraction from zaid
        patSpacing = re.compile(r'[\s\t]+')
        items = patSpacing.split(string)

        # ZAID
        pieces = items[0].split('.')
        # Try to identify library if present
        try:
            library = pieces[1]
        except IndexError:
            library = None

        # Identify element and isotope
        element = pieces[0][:-3]
        isotope = pieces[0][-3:]

        # identify fraction
        fraction = items[1]

        return cls(fraction, element, isotope, library)

    def to_text(self) -> str:
        """
         Get the zaid string ready for MCNP material card

        Returns
        -------
        str
            zaid string.

        """
        fraction = "{:.6E}".format(Decimal(self.fraction))
        if self.library is None:
            zaidname = self.element+self.isotope
        else:
            zaidname = self.element+self.isotope+'.'+self.library
        # Add INFO
        try:
            abundance = '%s' % float('%.5g' % float(self.ab))
        except ValueError:
            abundance = ''

        abundance = 'AB(%) '+abundance
        inline_comm = '    $ '+self.fullname
        args = (zaidname, fraction, inline_comm, abundance)

        return '{0:>15} {1:>18} {2:<12} {3:<10}'.format(*args)

    def get_fullname(self, libmanager: LibManager) -> str:
        """
        Get the formula name of the zaid (e.g. H1)

        Parameters
        ----------
        libmanager : libmanager.LibManager
            libmanager handling the libraries operations.

        Returns
        -------
        formula : str
            zaid formula name.

        """
        name, formula = libmanager.get_zaidname(self)

        return formula


#    def update_info(self,ab,fullname):
#        """
#        Update zaid info
#        """
#        self.additional_info['ab'] = ab
#        self.additional_info['fullname'] = fullname


class Element:

    def __init__(self, zaidList: list[Zaid]) -> None:
        """
        Generate an Element object starting from a list of zaids.
        It will collapse multiple instance of a zaid into a single one

        Parameters
        ----------
        zaidList : list
            list of zaids constituing the element.

        Returns
        -------
        None.

        Attributes
        ----------
        Z : str
            element str notation (AA) read from the zaids.
        zaids: list[Zaid]

        """
        zaids = {}
        for zaid in zaidList:
            # If already in dic sum the fractions
            if zaid.name in zaids.keys():
                zaids[zaid.name] = zaids[zaid.name]+zaid.fraction
            else:
                zaids[zaid.name] = zaid.fraction

        zaidList = []
        for name, fraction in zaids.items():
            zaidList.append(Zaid.from_string(name+' '+str(fraction)))

        self.Z = zaid.element
        self.zaids = zaidList

    def update_zaidinfo(self, libmanager: LibManager) -> None:
        """
        Update zaids infos through a libmanager. Info are the formula name and
        the abundance in the material.

        Parameters
        ----------
        libmanager : libmanager.LibManager
            libmanager handling the libraries operations.

        Returns
        -------
        None.

        """
        tot_fraction = 0
        for zaid in self.zaids:
            tot_fraction = tot_fraction + zaid.fraction

        for zaid in self.zaids:
            fullname = zaid.get_fullname(libmanager)
            ab = zaid.fraction/tot_fraction*100
#            zaid.update_info(ab,fullname)
            zaid.ab = ab
            zaid.fullname = fullname

    def get_fraction(self) -> float:
        """
        Get the sum of the fraction of the zaids composing the element

        Returns
        -------
        fraction : float
            element fraction.

        """
        fraction = 0
        for zaid in self.zaids:
            fraction = fraction+zaid.fraction

        return fraction


class SubMaterial:

    # init method for zaid
    def __init__(self, name: str, zaidList: list[Zaid],
                 elemList: list[Element] = None, header: str = None,
                 additional_keys: list[str] = None) -> None:
        """
        Generate a SubMaterial Object starting from a list of Zaid and
        eventually Elements list. Usually this kind of objects are generated
        directly reading a full material card, and rarely instanciated directly
        with the __init__ method.

        Parameters
        ----------
        name : str
            if the first submaterial, the name is the name of the material
            (e.g. m1).
        zaidList : list[Zaid]
            list of zaids composing the submaterial.
        elemList : list[Element], optional
            list of elements composing the submaterial. The default is None.
        header : str, optional
            Header of the submaterial. The default is None.
        additional_keys : list[str], optional
            list of additional keywords in the submaterial. The default is
            None.

        Returns
        -------
        None.

        Attributes
        ----------
        zaidList: list[Zaid]
            list of zaids in the sub-material
        elements: list[Element]
            list of elements in the sub-material
        header: str
            comment in the MCNP input file that is the header of the submat
        additional_keys: list[str]
            list of additional keys that may be part of the material

        """

        # List of zaids object of the submaterial
        self.zaidList = zaidList

        # Name of the material
        if name is not None:
            self.name = name.strip()  # Be sure to strip spaces
        else:
            self.name = None

        # List of elements in material
        if elemList is None:
            self._collapse_zaids()
        else:
            self.elements = elemList

        # Header of the submaterial
        self.header = header

        # Additional keys as plib,hlib etc.
        if additional_keys is None:
            additional_keys = []
        self.additional_keys = additional_keys

    @classmethod
    def from_text(cls, text: str) -> SubMaterial:
        """
        Generate a submaterial from MCNP input text

        Parameters
        ----------
        text : list[str]
            Original text of the MCNP input.

        Returns
        -------
        SubMaterial
            generated submaterial.

        """
        # Useful patterns
        patSpacing = re.compile(r'[\s\t]+')
        patComment = PAT_COMMENT
        patName = PAT_MAT
        searchHeader = True
        header = ''
        zaidList = []
        additional_keys_list = []
        for line in text:
            zaids = None
            additional_keys = None
            # Header MUST be at the top of the text block
            if searchHeader:
                # Get header
                if patComment.match(line) is None:
                    searchHeader = False
                    # Special treatment for first line
                    try:
                        name = patName.match(line).group()
                    except AttributeError:
                        # There is no material name
                        name = None

                    pieces = patSpacing.split(line)
                    if len(pieces) > 1:
                        # CASE1: only material name+additional spacing
                        if pieces[1] == '':
                            pass  # no more actions for this line
                        # CASE2: material name + zaids or only zaids
                        else:
                            if name is None:
                                start = 0
                            else:
                                start = patName.match(line).end()
                            zaids, additional_keys = _readLine(line[start:])
                    # CASE3: only material name and no spacing
                    else:
                        pass  # no more actions for this line
                else:
                    header = header+line
                    continue
            else:
                zaids, additional_keys = _readLine(line)

            if zaids is not None:
                zaidList.extend(zaids)

            if additional_keys is not None:
                additional_keys_list.extend(additional_keys)

        return cls(name, zaidList, elemList=None, header=header[:-1],
                   additional_keys=additional_keys_list)

    def _collapse_zaids(self) -> None:
        """
        Organize zaids into their elements and collapse mutiple istances

        Returns
        -------
        None.

        """
        elements = {}
        for zaid in self.zaidList:
            if zaid.element not in elements.keys():
                elements[zaid.element] = [zaid]
            else:
                elements[zaid.element].append(zaid)

        elemList = []
        for element_tag, zaids in elements.items():
            elemList.append(Element(zaids))

        self.elements = elemList

    def to_text(self) -> str:
        """
        Write to text in MNCP format the submaterial

        Returns
        -------
        str
            formatted submaterial text.

        """
        if self.header is not None:
            text = self.header+'\n'
        else:
            text = ''
        # if self.name is not None:
        #     text = text+'\n'+self.name
        if self.elements is not None:
            for elem in self.elements:
                for zaid in elem.zaids:
                    text = text+zaid.to_text()+'\n'
        else:
            for zaid in self.zaidList:
                text = text+zaid.to_text()+'\n'

        # Add additional keys
        if len(self.additional_keys) > 0:
            text = text+'\t'
            for key in self.additional_keys:
                text = text+' '+key

        return text.strip('\n')

    def translate(self, newlib: dict | str, lib_manager: LibManager) -> None:
        """
        This method implements the translation logic of JADE. All zaids are
        translated accordingly to the newlib specified.

        Parameters
        ----------
        newlib : dict | str
            There are a few ways that newlib can be provided:

            1) str (e.g. 31c), the new library to translate to will be the
            one indicated;

            2) dic (e.g. {'98c' : '99c', '31c: 32c'}), the new library is
            determined based on the old library of the zaid

            3) dic (e.g. {'98c': [list of zaids], '31c': [list of zaids]}),
            the new library to be used is explicitly stated depending
            on the zaidnum.
        lib_manager : LibManager
            Object handling libraries operation.

        Returns
        -------
        None.

        """
        newzaids = []
        for zaid in self.zaidList:
            # Implement the capability to translate to different libraries
            # depending on the starting one
            if type(newlib) == dict:
                # Check for which kind of dic it is
                if type(list(newlib.values())[0]) == str:
                    # The assignment is based on old lib
                    try:
                        newtag = newlib[zaid.library]
                    except KeyError:
                        # the zaid should have been assigned to a library
                        raise ValueError('''
 Zaid {} was not assigned to any library'''.format(zaid.name))

                else:
                    # The assignment is explicit, all libs need to be searched
                    newtag = None
                    zaidnum = zaid.element+zaid.isotope
                    for lib, zaids in newlib.items():
                        if zaidnum in zaids:
                            newtag = lib
                            break
                    # Check that a library has been actually found
                    if newtag is None:
                        # the zaid should have been assigned to a library
                        raise ValueError('''
 Zaid {} was not assigned to any library'''.format(zaid.name))
            else:
                newtag = newlib

            try:
                translation = lib_manager.convertZaid(zaid.element +
                                                      zaid.isotope, newtag)
            except ValueError:
                # No Available translation was found, ignore zaid
                # Only video warning, to propagate to the log would be too much
                print('  WARNING: no available translation was found for ' +
                      zaid.name+'.\n  The zaid has been ignored. ')
                continue

            # Check if it is  atomic or mass fraction
            if float(zaid.fraction) < 0:
                ref_mass = 0
                for key, item in translation.items():
                    ref_mass = ref_mass + item[1]*item[2]

                for key, item in translation.items():
                    fraction = str(item[1]*item[2]/ref_mass*zaid.fraction)
                    element = str(key)[:-3]
                    isotope = str(key)[-3:]
                    library = item[0]

                    newzaids.append(Zaid(fraction, element, isotope, library))

            else:
                for key, item in translation.items():
                    fraction = str(item[1]*zaid.fraction)
                    element = str(key)[:-3]
                    isotope = str(key)[-3:]
                    library = item[0]

                    newzaids.append(Zaid(fraction, element, isotope, library))

        self.zaidList = newzaids
        self._collapse_zaids()

    def _update_info(self, lib_manager: LibManager) -> None:
        """
        This methods allows to update the in-line comments for every zaids
        containing additional information

        Parameters
        ----------
        lib_manager : libmanager.LibManager
            Library manager for the conversion.

        Returns
        -------
        None.

        """
        self._collapse_zaids()  # To be sure to have adjourned elements

        for elem in self.elements:
            elem.update_zaidinfo(lib_manager)

        # TODO
        # Here the zaidlist of the submaterial should be adjourned or the next
        # collapse zaid will cancel the informations. If update info is used
        # as last operations there are no problems.

    def get_info(self, lib_manager: LibManager) -> tuple[pd.DataFrame]:
        """
        Returns DataFrame containing the different fractions of the elements
        and zaids

        Parameters
        ----------
        lib_manager : libmanager.LibManager
            Library manager for the conversion.

        Returns
        -------
        df_el : pd.DataFrame
            table of information of the submaterial on an elemental level.
        df_zaids : pd.DataFrame
            table of information of the submaterial on a zaid level.

        """
        # dic_element = {'Element': [], 'Fraction': []}
        # dic_zaids = {'Element': [], 'Zaid': [], 'Fraction': []}
        dic_element = {'Element': [], 'Fraction': []}
        dic_zaids = {'Element': [], 'Isotope': [], 'Fraction': []}
        for elem in self.elements:
            fraction = elem.get_fraction()
            # dic_element['Element'].append(elem.Z)
            dic_element['Fraction'].append(fraction)
            for zaid in elem.zaids:
                fullname = zaid.get_fullname(lib_manager)
                elementname = fullname.split('-')[0]
                # dic_zaids['Element'].append(elem.Z)
                # dic_zaids['Zaid'].append(zaid.isotope)
                dic_zaids['Element'].append(elementname)
                dic_zaids['Isotope'].append(fullname + ' [' + str(zaid.element)
                                            + str(zaid.isotope)+']')
                dic_zaids['Fraction'].append(zaid.fraction)

            dic_element['Element'].append(elementname)

        df_el = pd.DataFrame(dic_element)
        df_zaids = pd.DataFrame(dic_zaids)

        return df_el, df_zaids

    def scale_fractions(self, norm_factor: float) -> None:
        """
        Scale the zaids fractions using a normalizing factor

        Parameters
        ----------
        norm_factor : float
            scaling factor.

        Returns
        -------
        None.

        """
        for zaid in self.zaidList:
            zaid.fraction = zaid.fraction*norm_factor

        self._collapse_zaids()


# Support function for Submaterial
def _readLine(string: str) -> tuple[list[Zaid], list[str] | None]:
    patSpacing = re.compile(r'[\s\t]+')
    patComment = re.compile(r'\$')
    patnumber = re.compile(r'\d+')

    pieces = patSpacing.split(string)
    # kill first piece if it is void
    if pieces[0] == '':
        del pieces[0]
    # kill last piece if it is void
    if pieces[-1] == '':
        del pieces[-1]

    # kill comment section
    i = 0
    for i, piece in enumerate(pieces):
        if patComment.match(pieces[i]) is not None:
            del pieces[i:]
            break

    i = 0
    zaids = []
    additional_keys = None
    while True:
        try:
            # Check if it is zaid or keyword
            if patnumber.match(pieces[i]) is None or pieces[i] == '':
                additional_keys = pieces[i:]
                break
            else:
                zaidstring = pieces[i]+' '+pieces[i+1]
                zaid = Zaid.from_string(zaidstring)
                zaids.append(zaid)

            i = i+2

        except IndexError:
            break

    return zaids, additional_keys


class Material:

    def __init__(self, zaids: list[Zaid], elem: list[Element], name: str,
                 submaterials: list[SubMaterial] = None, mx_cards: list = None,
                 header: str = None) -> None:
        """
        Object representing an MCNP material

        Parameters
        ----------
        zaids : list[zaids]
            zaids composing the material.
        elem : list[elem]
            elements composing the material.
        name : str
            name of the material (e.g. m1).
        submaterials : list[Submaterials], optional
            list of submaterials composing the material. The default is None.
        mx_cards : list, optional
            list of mx_cards in the material if present. The default is None.
        header : str, optional
            material header. The default is None.

        Returns
        -------
        None.

        Attributes
        ----------
        all __init__ parameters are stored as attributes.

        """

        self.zaids = zaids
        self.elem = elem
        self.submaterials = submaterials
        self.name = name.strip()
        if mx_cards is None:
            self.mx_cards = []
        else:
            self.mx_cards = mx_cards
        self.header = header

        # Adjust the submaterial and headers reading
        try:
            # The first submaterial header is actually the material header.
            # If submat is void it has to be deleted (only header), otherwise
            # it means it has no header
            submat = submaterials[0]
            if len(submat.zaidList) == 0:  # Happens in reading from text
                # self.header = submat.header
                del self.submaterials[0]
            # else:
            #     self.submaterials[0].header = None
        except IndexError:
            self.header = None

    @classmethod
    def from_text(cls, text: list[str]) -> Material:
        """
        Create a material from MCNP formatted text

        Parameters
        ----------
        text : list[str]
            MCNP formatted text.

        Returns
        -------
        matreader.Material
            material object created.

        """
        # split the different submaterials
        patC = PAT_COMMENT
        pat_matHeader = PAT_MAT
        inHeader = True
        subtext = []
        submaterials = []

        for line in text:
            checkComment = patC.match(line)
            checkHeaderMat = pat_matHeader.match(line)

            if checkHeaderMat is not None:
                header = ''.join(subtext)
                subtext = []

            if inHeader:
                subtext.append(line)
                if checkComment is None:  # The end of the header is found
                    inHeader = False
            else:
                if checkComment is None:  # Still in the material
                    subtext.append(line)
                else:  # a new header starts
                    submaterials.append(SubMaterial.from_text(subtext))
                    inHeader = True
                    subtext = [line]

        submaterials.append(SubMaterial.from_text(subtext))

        return cls(None, None, submaterials[0].name, submaterials=submaterials,
                   header=header)

    def to_text(self) -> str:
        """
        Write the material to MCNP formatted text

        Returns
        -------
        str
            MCNP formatte text representing the material.

        """
        if self.header is not None:
            text = self.header.strip('\n')+'\n'+self.name.upper().strip('\n')
        else:
            text = self.name.upper()
        if self.submaterials is not None:
            for submaterial in self.submaterials:
                text = text+'\n'+submaterial.to_text()
            # Add mx cards
            for mx in self.mx_cards:
                for line in mx.lines:
                    line = line.strip('\n')
                    text = text+'\n'+line.upper()
        else:
            text = '  Not supported yet, generate submaterials first'
            pass  # TODO

        return text.strip('\n')

    def translate(self, newlib: dict | str, lib_manager: LibManager) -> None:
        """
        This method allows to translate all submaterials to another library

        Parameters
        ----------
        newlib : dict or str
            There are a few ways that newlib can be provided:

            1) str (e.g. 31c), the new library to translate to will be the
            one indicated;

            2) dic (e.g. {'98c' : '99c', '31c: 32c'}), the new library is
            determined based on the old library of the zaid

            3) dic (e.g. {'98c': [list of zaids], '31c': [list of zaids]}),
            the new library to be used is explicitly stated depending
            on the zaidnum.
        lib_manager : libmanager.LibManager
            object handling all libraries operations.

        Returns
        -------
        None.
        """
        for submat in self.submaterials:
            submat.translate(newlib, lib_manager)

        self._update_info(lib_manager)

    def get_tot_fraction(self) -> float:
        """
        Returns the total material fraction
        """
        fraction = 0
        for submat in self.submaterials:
            for zaid in submat.zaidList:
                fraction = fraction+zaid.fraction

        return fraction

    def add_mx(self, mx_cards: list) -> None:
        """
        Add a list of mx_cards to the material
        """
        self.mx_cards.append(mx_cards)

    def _update_info(self, lib_manager: LibManager) -> None:
        """
        This methods allows to update the in-line comments for every zaids
        containing additional information

        lib_manager: (LibManager) Library manager for the conversion
        """
        for submaterial in self.submaterials:
            submaterial._update_info(lib_manager)

    def switch_fraction(self, ftype: str, lib_manager: LibManager,
                        inplace: bool = True) -> list[SubMaterial]:
        """
        Switch between atom or mass fraction for the material card.
        If the material is already switched the command is ignored.

        Parameters
        ----------
        ftype : str
            Either 'mass' or 'atom' to chose the type of switch.
        lib_manager : libmanager.LibManager
            Handles zaid data.
        inplace : bool
            if True the densities of the isotopes are changed inplace,
            otherwise a copy of the material is provided. DEFAULT is True

        Raises
        ------
        KeyError
            if ftype is not either 'atom' or 'mass'.

        Returns
        -------
        submaterials : list[SubMaterial]
            list of the submaterials where fraction have been switched

        """
        # Get total fraction
        totf = self.get_tot_fraction()
        new_submats = []

        if ftype == 'atom':  # mass2atom switch
            if totf < 0:  # Check if the switch must be effectuated
                # x_n = (x_m/m)/sum(x_m/m)
                # get sum(x_m/m)
                norm = 0
                for submat in self.submaterials:
                    for zaid in submat.zaidList:
                        atom_mass = lib_manager.get_zaid_mass(zaid)
                        norm = norm + (-1*zaid.fraction/atom_mass)

                for submat in self.submaterials:
                    new_zaids = []
                    new_submat = copy.deepcopy(submat)
                    for zaid in submat.zaidList:
                        atom_mass = lib_manager.get_zaid_mass(zaid)
                        if inplace:
                            zaid.fraction = (-1*zaid.fraction/atom_mass)/norm
                        else:
                            newz = copy.deepcopy(zaid)
                            newz.fraction = (-1*zaid.fraction/atom_mass)/norm
                            new_zaids.append(newz)
                    new_submat.zaidList = new_zaids
                    new_submat._update_info(lib_manager)
                    new_submats.append(new_submat)
            else:
                new_submats = self.submaterials

        elif ftype == 'mass':  # atom2mass switch
            if totf > 0:  # Check if the switch must be effectuated
                # x_n = (x_m*m)/sum(x_m*m)
                # get sum(x_m*m)
                norm = 0
                for submat in self.submaterials:
                    for zaid in submat.zaidList:
                        atom_mass = lib_manager.get_zaid_mass(zaid)
                        norm = norm + (zaid.fraction*atom_mass)

                for submat in self.submaterials:
                    new_zaids = []
                    new_submat = copy.deepcopy(submat)
                    for zaid in submat.zaidList:
                        atom_mass = lib_manager.get_zaid_mass(zaid)
                        if inplace:
                            zaid.fraction = (-1*zaid.fraction*atom_mass)/norm
                        else:
                            newz = copy.deepcopy(zaid)
                            newz.fraction = (-1*zaid.fraction*atom_mass)/norm
                            new_zaids.append(newz)
                    new_submat.zaidList = new_zaids
                    new_submat._update_info(lib_manager)
                    new_submats.append(new_submat)
            else:
                new_submats = self.submaterials

        else:
            raise KeyError(ftype+' is not a valid key error [atom, mass]')

        self._update_info(lib_manager)

        return new_submats


class MatCardsList(Sequence):

    def __init__(self, materials: list[Material]) -> None:
        """
        Object representing the list of materials included in an MCNP input.
        This class is a child of the Sequence base class.

        Parameters
        ----------
        materials : list[Material]
            list of materials.

        Returns
        -------
        None.

        Attributes
        ----------
        materials: list[Material]
            list of the material objects that compose the material section
        matdic: dict[str, Material]
            dictionary of the material objects that compose the material
            section. This attribute is used mostly as internal object, the
            recommended way to access materials is to use directly the
            "dictionary" capabilities of the MatCardsList object.

        Examples
        --------

        >>> from f4enix.input.inputAPI import MatCardsList
        ... # initialize from file
        ... materials = MatCardsList.from_input('inputfile.i')
        ... # get a specific material
        ... mat1 = materials['m1']

        """
        self.materials = materials
        # Build also the dictionary
        self.matdic = self._compute_dic()

    def __len__(self) -> int:
        return len(self.materials)

    def __repr__(self) -> str:
        return str(self.matdic)

    def __str__(self) -> str:
        return str(self.matdic)

    def __getitem__(self, key: str | int) -> Material:
        if type(key) is int:
            return self.materials[key]
        else:
            return self.matdic[key.upper()]

    def append(self, material: Material) -> None:
        self.materials.append(material)
        self.matdic = self._compute_dic()

    def remove(self, item: Material) -> None:
        self.materials.remove(item)  # TODO this should get the key instead
        self.matdic = self._compute_dic()

    def _compute_dic(self) -> dict[str, Material]:
        matdic = {}
        for material in self.materials:
            matdic[material.name.upper()] = material

        return matdic

    @classmethod
    def from_input(cls, inputfile: os.PathLike) -> MatCardsList:
        """
        This method use the numjuggler parser to help identify the mcards in
        the input. Then the mcards are parsed using the classes defined in this
        module

        Parameters
        ----------
        inputfile : os.PathLike
            MCNP input file containing the material section.

        Returns
        -------
        MatCardsList
            new material card list generated.

        """
        matPat = PAT_MAT
        mxPat = PAT_MX
        commentPat = PAT_COMMENT
        # Using parser the data cards are extracted from the input.
        # Comment section are interpreted as cards by the parser
        with suppress_stdout():
            # Suppress output from tab replacing
            cards = par.get_cards_from_input(inputfile)
            cardsDic = par.get_blocks(cards)
        datacards = cardsDic[5]

        materials = []
        previous_lines = ['']
        mx_cards = []
        mx_found = False

        for datacard in datacards:
            lines = datacard.lines

            # Check if it is a material card
            if matPat.match(lines[0]) is not None:
                # Check if previous card is the header
                if commentPat.match(previous_lines[0]):
                    previous_lines.extend(lines)
                    material = Material.from_text(previous_lines)
                else:
                    material = Material.from_text(lines)

                materials.append(material)

            # Check if the current is an mx cards
            if mxPat.match(lines[0]) is not None:
                mx_cards.append(lines)
                mx_found = True

            # If not Add mx cards if previous one was an mx
            elif mx_found:
                materials[-1].add_mx(mx_cards)
                mx_cards = []
                mx_found = False

            else:
                mx_found = False

            previous_lines = lines

        # If material is last datacard
        if mx_found:
            materials[-1].add_mx(mx_cards)

        return cls(materials)

    def to_text(self) -> str:
        """
        return text of the material cards in order

        Returns
        -------
        str
            material card list MCNP formatted text.

        """
        text = ''
        for material in self.materials:
            text = text+'\n'+material.to_text()

        return (text.strip('\n'))

    def translate(self, newlib: str | dict, lib_manager: LibManager) -> None:
        """
        This method allows to translate the material cards to another library.
        The zaid are collapsed again to get the new elements

        Parameters
        ----------
        newlib : dict or str
            There are a few ways that newlib can be provided:

            1) str (e.g. 31c), the new library to translate to will be the
            one indicated;

            2) dic (e.g. {'98c' : '99c', '31c: 32c'}), the new library is
            determined based on the old library of the zaid

            3) dic (e.g. {'98c': [list of zaids], '31c': [list of zaids]}),
            the new library to be used is explicitly stated depending
            on the zaidnum.

        lib_manager : libmanager.LibManager
            Library manager for the conversion.

        Returns
        -------
        None.

        """
        for material in self.materials:
            material.translate(newlib, lib_manager)

            # Rebuild elements
            for submat in material.submaterials:
                submat._collapse_zaids()

    def update_info(self, lib_manager: LibManager) -> None:
        """
        This methods allows to update the in-line comments for every zaids
        containing additional information

        Parameters
        ----------
        lib_manager : libmanager.Libmanager
            Library manager for the conversion.

        Returns
        -------
        None.

        """
        for mat in self.materials:
            mat._update_info(lib_manager)

    def get_info(self, lib_manager: LibManager, zaids: bool = False,
                 complete: bool = True) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Get the material informations in terms of fraction and composition
        of the material card

        Parameters
        ----------
        lib_manager : libmanager.LibManager
            To handle element name recovering.
        zaids : bool, optional
            Consider or not the zaid level. The default is False.
        complete: bool, optional
            If True both the atom and mass fraction are given in the raw
            table. The default is True.

        Returns
        -------
        df : pd.DataFrame
            Raw infos on the fractions.
        df_elem : pd.DataFrame
            processed info for the element: normalized fraction added both for
            material and submaterial.

        """
        infos = []
        complete_infos = []
        for mat in self.materials:
            submats_atom = mat.switch_fraction('atom', lib_manager,
                                               inplace=False)
            submats_mass = mat.switch_fraction('mass', lib_manager,
                                               inplace=False)
            i = 0
            for submat, submat_a, submat_m in zip(mat.submaterials,
                                                  submats_atom, submats_mass):
                dic_el, dic_zaids = submat.get_info(lib_manager)
                dic_el_a, dic_zaids_a = submat_a.get_info(lib_manager)
                dic_el_m, dic_zaids_m = submat_m.get_info(lib_manager)

                if zaids:
                    dic = dic_zaids
                    dic_a = dic_zaids_a
                    dic_m = dic_zaids_m
                else:
                    dic = dic_el
                    dic_a = dic_el_a
                    dic_m = dic_el_m

                dic['Material'] = mat.name
                dic['Submaterial'] = i+1
                infos.append(dic)

                c_dic = copy.deepcopy(dic)
                c_dic['Atom Fraction'] = dic_a['Fraction']
                c_dic['Mass Fraction'] = dic_m['Fraction']
                complete_infos.append(c_dic)

                i = i+1

        df = pd.concat(infos)
        df_complete = pd.concat(complete_infos)
        del df_complete['Fraction']

        if zaids:
            df.set_index(['Material', 'Submaterial', 'Element', 'Isotope'],
                         inplace=True)
            df_complete.set_index(['Material', 'Submaterial', 'Element',
                                   'Isotope'], inplace=True)

        else:
            df.set_index(['Material', 'Submaterial', 'Element'], inplace=True)
            df_complete.set_index(['Material', 'Submaterial', 'Element'],
                                  inplace=True)

        # Additional df containing normalized element fraction of submaterial
        # and material

        # Get total fractions
        df_elem = df.groupby(['Material', 'Submaterial', 'Element']).sum()
        df_sub = df.groupby(['Material', 'Submaterial']).sum()
        df_mat = df.groupby(['Material']).sum()

        # Compute percentages
        sub_percentage = []
        mat_percentage = []
        for idx, row in df_elem.iterrows():
            matID = idx[0]
            elemID = idx[1]
            sub_percentage.append(row['Fraction'] /
                                  df_sub['Fraction'].loc[(matID, elemID)])
            mat_percentage.append(row['Fraction'] /
                                  df_mat['Fraction'].loc[matID])

        df_elem['Sub-Material Fraction'] = sub_percentage
        df_elem['Material Fraction'] = mat_percentage

        if complete:
            return df_complete, df_elem
        else:
            return df, df_elem

    def generate_material(self, materials: list[str],
                          percentages: list[float],
                          newlib: str,
                          libmanager: LibManager,
                          fractiontype='atom',
                          mat_name='M1') -> Material:
        """
        Starting from an MCNP input, materials contained in its material list
        can be used to generate a new material combining them.

        Parameters
        ----------
        materials : list[str]
            list of materials to mix (e.g. ['m1', 'M2']).
        percentages : list[float]
            percentages associated to the source materials in the new materials
            (e.g. [0.1, 0.9)]. Their are intended as atom or mass fraction
            depending on the fractiontype that is specified.
        newlib : str
            library for the new material.
        fractiontype : str, optional
            type of fraction to use in the new material (either 'atom' or
            'mass'. The default is 'atom'.
        mat_name : str, optional
            Material card name of the new generated material. the default is 
            'M1'
        Returns
        -------
        Material
            Newly created material

        """

        if re.match(r'^M\d{1,7}$', mat_name) is None:
            print('\nMaterial name not valid, set to M1\n')
            mat_name = 'M1'
 
        # Translate to requested lib
        self.translate(newlib, libmanager)

        # Collect all submaterials
        submaterials = []
        main_header = ''
        for materialname, percentage in zip(materials, percentages):
            materialname = materialname.upper()
            percentage_str = str(round(float(percentage)*100, 2))+'%'
            main_header = (
                '{}C Material: {} Percentage: {} ({})\n'.format(main_header,
                                                                materialname,
                                                                percentage_str,
                                                                fractiontype))
            material = copy.deepcopy(self[materialname])
            # Ensure materials have the requested fraction type
            material.switch_fraction(fractiontype, libmanager)

            # Scale fractions
            totfraction = material.get_tot_fraction()
            current_submaterials = []
            for j, submat in enumerate(material.submaterials):
                # normalized & scaled
                norm_factor = float(percentage)/totfraction
                if fractiontype == 'mass':
                    norm_factor = -norm_factor
                submat.scale_fractions(norm_factor)
                submat._update_info(libmanager)
                # Add info to the header in order to back-trace the generation
                submat.header = (
                    'C {}, submaterial {}\n{}'.format(materialname,
                                                      j+1,
                                                      submat.header))
                # Drop additional keys if present
                submat.additional_keys = []
                current_submaterials.append(submat)

            # Change the header of the first submaterial to include the mat. 1
            new_sub_header = (material.header +
                              current_submaterials[0].header).strip('\n')
            current_submaterials[0].header = new_sub_header
            submaterials.extend(current_submaterials)

        # Generate new material and matlist
        newmat = Material(None, None, mat_name, submaterials=submaterials,
                          header=main_header)

        return newmat


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
